# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pickle
from typing import Dict

import numpy as np

from ... import opcodes as OperandDef
from ...serialize import ValueType, KeyField, StringField, BytesField, TupleField
from ...filesystem import get_fs, FSMap
from ...tiles import TilesError
from ...utils import check_chunks_unknown_shape
from .core import TensorDataStore


class ZarrOptions(object):
    def __init__(self, options: Dict):
        self._options = options

    def todict(self):
        return self._options

    @staticmethod
    def _stringfy(v):
        return pickle.dumps(v) if not isinstance(v, str) else v

    def __mars_tokenize__(self):
        return list(self._options.keys()), \
               list(self._stringfy(v) for v in self._options.values())

    def __getstate__(self):
        return self._options

    def __setstate__(self, state):
        self._options = state


class TensorToZarrDataStore(TensorDataStore):
    _op_type_ = OperandDef.TENSOR_STORE_ZARR

    _input = KeyField('input')
    _path = StringField('path')
    _group = StringField('group')
    _dataset = StringField('dataset')
    _zarr_options = BytesField('zarr_options', on_serialize=pickle.dumps,
                               on_deserialize=pickle.loads)
    _axis_offsets = TupleField('axis_offsets', ValueType.int32)

    def __init__(self, path=None, group=None, dataset=None, zarr_options=None,
                 axis_offsets=None, **kw):
        super().__init__(_path=path, _group=group, _dataset=dataset,
                         _zarr_options=zarr_options, _axis_offsets=axis_offsets, **kw)

    @property
    def path(self):
        return self._path

    @property
    def group(self):
        return self._group

    @property
    def dataset(self):
        return self._dataset

    @property
    def zarr_options(self):
        return self._zarr_options

    @property
    def axis_offsets(self):
        return self._axis_offsets

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._input = self._inputs[0]

    @classmethod
    def tile(cls, op):
        import zarr

        check_chunks_unknown_shape(op.inputs, TilesError)
        in_tensor = op.input

        # create dataset
        fs = get_fs(op.path, None)
        path = op.path
        if op.group is not None:
            path += '/' + op.group
        fs_map = FSMap(path, fs)
        zarr.open(fs_map, 'w', path=op.dataset,
                  dtype=in_tensor.dtype, shape=in_tensor.shape,
                  chunks=tuple(max(ns) for ns in in_tensor.nsplits),
                  **op.zarr_options.todict())

        cum_nsplits = [[0] + np.cumsum(ns).tolist() for ns in in_tensor.nsplits]
        out_chunks = []
        for chunk in in_tensor.chunks:
            chunk_op = op.copy().reset_key()
            chunk_op._axis_offsets = \
                tuple(cs[i] for i, cs in zip(chunk.index, cum_nsplits))
            out_chunks.append(chunk_op.new_chunk([chunk], shape=(0,) * chunk.ndim,
                                                 index=chunk.index))

        new_op = op.copy()
        out = op.outputs[0]
        nsplits = tuple((0,) * len(ns) for ns in in_tensor.nsplits)
        return new_op.new_tensors(op.inputs, shape=out.shape, order=out.order,
                                  nsplits=nsplits, chunks=out_chunks)

    @classmethod
    def execute(cls, ctx, op):
        import zarr

        fs = get_fs(op.path, None)
        fs_map = FSMap(op.path, fs)

        group = zarr.Group(store=fs_map, path=op.group)
        array = group[op.dataset]

        to_store = ctx[op.inputs[0].key]
        axis_offsets = op.axis_offsets
        shape = to_store.shape

        array[tuple(slice(offset, offset + size)
              for offset, size
              in zip(axis_offsets, shape))] = to_store

        ctx[op.outputs[0].key] = np.empty((0,) * to_store.ndim,
                                          dtype=to_store.dtype)


def tozarr(path, x, group=None, dataset=None, **zarr_options):
    import zarr

    if isinstance(path, zarr.Array):
        arr = path
        if isinstance(arr.store, FSMap):
            root = arr.store.root
            path, dataset = root.rsplit('/', 1)
        else:
            path = arr.store.path
            if '/' in arr.path and group is None:
                group = arr.path.rsplit('/', 1)[0]
            dataset = arr.basename
            if not dataset:
                path, dataset = path.rsplit('/', 1)
        for attr in ['compressor', 'filters']:
            if getattr(arr, attr):
                zarr_options[attr] = getattr(arr, attr)
    elif isinstance(path, str):
        if dataset is None:
            path, dataset = path.rsplit('/', 1)
    else:
        raise TypeError('`path` passed has wrong type, '
                        'expect str, or zarr.Array'
                        f'got {type(path)}')

    op = TensorToZarrDataStore(path=path, group=group, dataset=dataset,
                               zarr_options=ZarrOptions(zarr_options))
    return op(x)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
try:
    import tiledb
except (ImportError, OSError):  # pragma: no cover
    tiledb = None


def get_tiledb_schema_from_tensor(tensor, tiledb_ctx, nsplits, **kw):
    from ..core import TensorOrder

    ctx = tiledb_ctx

    dims = []
    for d in range(tensor.ndim):
        extent = tensor.shape[d]
        domain = (0, extent - 1)
        tile = max(nsplits[d])
        dims.append(tiledb.Dim(name="", domain=domain, tile=tile, dtype=np.int64, ctx=ctx))
    dom = tiledb.Domain(*dims, ctx=ctx)
    att = tiledb.Attr(ctx=ctx, dtype=tensor.dtype)
    cell_order = 'C' if tensor.order == TensorOrder.C_ORDER else 'F'
    return tiledb.ArraySchema(ctx=ctx, domain=dom, attrs=(att,),
                              sparse=tensor.issparse(), cell_order=cell_order, **kw)


def check_tiledb_array_with_tensor(tensor, tiledb_array):
    if tensor.ndim != tiledb_array.ndim:
        # ndim
        raise ValueError('ndim of TileDB Array to store is different to tensor, '
                         f'expect {tensor.ndim}, got {tiledb_array.ndim}')
    if tensor.shape != tiledb_array.shape:
        # shape
        raise ValueError('shape of TileDB Array to store is different to tensor, '
                         f'expect {tensor.shape}, got {tiledb_array.shape}')
    if tensor.dtype != tiledb_array.attr(0).dtype:
        # dtype
        raise ValueError('dtype of TileDB Array to store is different to tensor, '
                         f'expect {tensor.dtype}, got {tiledb_array.domain.dtype}')



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd
from pandas.api.types import is_list_like

from ... import opcodes
from ...config import options
from ...operands import OperandStage
from ...serialize import KeyField, AnyField, BoolField, StringField, Int32Field
from ...tiles import TilesError
from ...utils import check_chunks_unknown_shape, ceildiv, get_shuffle_input_keys_idxes, lazy_import
from ..initializer import DataFrame as asdataframe
from ..operands import DataFrameMapReduceOperand, DataFrameOperandMixin, \
    DataFrameShuffleProxy, OutputType
from ..merge import DataFrameConcat
from ..utils import parse_index, hash_dataframe_on, gen_unknown_index_value, standardize_range_index

cudf = lazy_import('cudf', globals=globals())


class DataFrameDropDuplicates(DataFrameMapReduceOperand, DataFrameOperandMixin):
    _op_type_ = opcodes.DROP_DUPLICATES

    _input = KeyField('input')
    _subset = AnyField('subset')
    _keep = AnyField('keep')
    _ignore_index = BoolField('ignore_index')
    _method = StringField('method')
    # subset chunk, used for method 'subset_tree'
    _subset_chunk = KeyField('subset_chunk')
    # shuffle phase, used in shuffle method
    _shuffle_size = Int32Field('shuffle_size')
    _shuffle_phase = StringField('shuffle_phase')

    def __init__(self, subset=None, keep=None, ignore_index=None,
                 output_types=None, method=None, subset_chunk=None,
                 shuffle_size=None, shuffle_phase=None, shuffle_key=None, **kw):
        super().__init__(_subset=subset, _keep=keep, _ignore_index=ignore_index,
                         _output_types=output_types, _method=method,
                         _subset_chunk=subset_chunk,
                         _shuffle_size=shuffle_size,
                         _shuffle_phase=shuffle_phase,
                         _shuffle_key=shuffle_key, **kw)

    @property
    def input(self):
        return self._input

    @property
    def subset(self):
        return self._subset

    @property
    def keep(self):
        return self._keep

    @property
    def ignore_index(self):
        return self._ignore_index

    @property
    def method(self):
        return self._method

    @property
    def subset_chunk(self):
        return self._subset_chunk

    @property
    def shuffle_size(self):
        return self._shuffle_size

    @property
    def shuffle_phase(self):
        return self._shuffle_phase

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._input = self._inputs[0]
        if self._subset_chunk is not None:
            self._subset_chunk = self._inputs[1]

    def __call__(self, inp, inplace=False):
        self._output_types = inp.op.output_types
        params = inp.params
        if self._ignore_index:
            params['index_value'] = parse_index(pd.RangeIndex(-1))
        else:
            params['index_value'] = gen_unknown_index_value(
                params['index_value'], self._keep, self._subset, type(self).__name__)
        shape_list = list(params['shape'])
        shape_list[0] = np.nan
        params['shape'] = tuple(shape_list)

        ret = self.new_tileable([inp], kws=[params])
        if inplace:
            inp.data = ret.data
        return ret

    @classmethod
    def _tile_one_chunk(cls, op: "DataFrameDropDuplicates"):
        inp = op.input
        out = op.outputs[0]

        chunk_op = op.copy().reset_key()
        chunk_op._method = None
        chunk_params = out.params
        chunk_params['index'] = (0,) * out.ndim
        chunk_shape_list = list(chunk_params['shape'])
        chunk_shape_list[0] = np.nan
        chunk_params['shape'] = tuple(chunk_shape_list)
        in_chunk = inp.chunks[0]
        chunk_params['index_value'] = gen_unknown_index_value(
            chunk_params['index_value'], in_chunk)
        chunk = chunk_op.new_chunk([in_chunk], kws=[chunk_params])

        params = out.params
        params['chunks'] = [chunk]
        params['nsplits'] = tuple((s,) for s in chunk.shape)
        new_op = op.copy()
        return new_op.new_tileables([inp], kws=[params])

    @classmethod
    def _gen_map_chunks(cls, op: "DataFrameDropDuplicates", inp, method, **kw):
        chunks = inp.chunks
        map_chunks = []
        for c in chunks:
            chunk_op = op.copy().reset_key()
            chunk_op._method = method
            chunk_op._stage = OperandStage.map
            for k, v in kw.items():
                setattr(chunk_op, k, v)
            chunk_params = c.params
            chunk_shape_list = list(chunk_params['shape'])
            chunk_shape_list[0] = np.nan
            chunk_params['shape'] = tuple(chunk_shape_list)
            chunk_params['index_value'] = \
                gen_unknown_index_value(chunk_params['index_value'], c)
            map_chunks.append(chunk_op.new_chunk([c], kws=[chunk_params]))
        return map_chunks

    @classmethod
    def _gen_tree(cls, op: "DataFrameDropDuplicates", inp, method):
        combine_size = options.combine_size
        new_chunks = cls._gen_map_chunks(op, inp, method)
        while len(new_chunks) > 1:
            out_chunk_size = ceildiv(len(new_chunks), combine_size)
            out_chunks = []
            for i in range(out_chunk_size):
                in_chunks = new_chunks[i * combine_size: (i + 1) * combine_size]
                if in_chunks[0].ndim == 2:
                    kw = dict(dtypes=in_chunks[0].dtypes)
                else:
                    kw = dict(dtype=in_chunks[0].dtype)
                concat_chunk = DataFrameConcat(
                    output_types=in_chunks[0].op.output_types).new_chunk(
                    in_chunks, **kw)
                chunk_op = op.copy().reset_key()
                chunk_op._method = method
                chunk_op._stage = \
                    OperandStage.combine if out_chunk_size > 1 else OperandStage.agg
                params = in_chunks[0].params
                chunk_shape_list = list(params['shape'])
                chunk_shape_list[0] = sum(c.shape[0] for c in in_chunks)
                params['shape'] = tuple(chunk_shape_list)
                chunk_index_list = list(params['index'])
                chunk_index_list[0] = i
                params['index'] = tuple(chunk_index_list)
                params['index_value'] = gen_unknown_index_value(
                    params['index_value'], *in_chunks)
                out_chunks.append(chunk_op.new_chunk([concat_chunk], kws=[params]))
            new_chunks = out_chunks

        return new_chunks

    @classmethod
    def _tile_tree(cls, op: "DataFrameDropDuplicates", inp):
        out = op.outputs[0]

        params = out.params
        params['chunks'] = cls._gen_tree(op, inp, 'tree')
        params['nsplits'] = ((np.nan,),) + inp.nsplits[1:]
        new_op = op.copy()
        return new_op.new_tileables([inp], kws=[params])

    @classmethod
    def _tile_subset_tree(cls, op: "DataFrameDropDuplicates", inp):
        # subset is available for DataFrame only
        inp = asdataframe(inp)
        out = op.outputs[0]
        subset = op.subset
        if subset is None:
            subset = inp.dtypes.index.tolist()
        # select subset first
        subset_df = inp[subset]._inplace_tile()
        # tree aggregate subset
        subset_chunk = cls._gen_tree(op, subset_df, 'subset_tree')[0]

        out_chunks = []
        for c in inp.chunks:
            chunk_op = op.copy().reset_key()
            chunk_op._method = 'subset_tree'
            chunk_op._subset_chunk = subset_chunk
            chunk_params = c.params
            chunk_shape_list = list(chunk_params['shape'])
            chunk_shape_list[0] = np.nan
            chunk_params['shape'] = tuple(chunk_shape_list)
            out_chunks.append(chunk_op.new_chunk([c, subset_chunk],
                                                 kws=[chunk_params]))

        new_op = op.copy()
        params = out.params
        params['chunks'] = out_chunks
        params['nsplits'] = ((np.nan,) * len(inp.chunks),) + inp.nsplits[1:]
        return new_op.new_tileables([inp], kws=[params])

    @classmethod
    def _tile_shuffle(cls, op: "DataFrameDropDuplicates", inp):
        out = op.outputs[0]

        map_chunks = cls._gen_map_chunks(op, inp, 'shuffle',
                                         _shuffle_size=inp.chunk_shape[0])
        proxy_chunk = DataFrameShuffleProxy(
            output_types=map_chunks[0].op.output_types).new_chunk(map_chunks, shape=())
        reduce_chunks = []
        for i in range(len(map_chunks)):
            reduce_op = op.copy().reset_key()
            reduce_op._method = 'shuffle'
            reduce_op._stage = OperandStage.reduce
            reduce_op._shuffle_phase = 'drop_duplicates'
            reduce_op._shuffle_key = str(i)
            reduce_op._shuffle_size = inp.chunk_shape[0]
            reduce_chunk_params = map_chunks[0].params
            reduce_chunk_params['index'] = (i,) + reduce_chunk_params['index'][1:]
            reduce_chunks.append(
                reduce_op.new_chunk([proxy_chunk], kws=[reduce_chunk_params]))

        put_back_proxy_chunk = DataFrameShuffleProxy(
            output_types=map_chunks[0].op.output_types).new_chunk(reduce_chunks, shape=())
        put_back_chunks = []
        for i in range(len(map_chunks)):
            put_back_op = op.copy().reset_key()
            put_back_op._method = 'shuffle'
            put_back_op._stage = OperandStage.reduce
            put_back_op._shuffle_phase = 'put_back'
            put_back_op._shuffle_key = str(i)
            put_back_chunk_params = map_chunks[0].params
            put_back_chunk_params['index'] = (i,) + put_back_chunk_params['index'][1:]
            put_back_chunks.append(
                put_back_op.new_chunk([put_back_proxy_chunk], kws=[put_back_chunk_params]))

        if op.ignore_index:
            put_back_chunks = standardize_range_index(put_back_chunks)

        new_op = op.copy()
        params = out.params
        params['chunks'] = put_back_chunks
        params['nsplits'] = ((np.nan,) * len(put_back_chunks),) + inp.nsplits[1:]
        return new_op.new_tileables([inp], kws=[params])

    @classmethod
    def tile(cls, op: "DataFrameDropDuplicates"):
        inp = op.input

        if len(inp.chunks) == 1:
            # one chunk
            return cls._tile_one_chunk(op)

        if inp.ndim == 2 and inp.chunk_shape[1] > 1:
            check_chunks_unknown_shape([inp], TilesError)
            inp = inp.rechunk({1: inp.shape[1]})._inplace_tile()

        default_tile = cls._tile_tree

        if op.method == 'auto':
            # if method == 'auto', pick appropriate method
            if np.isnan(inp.shape[0]) or op.subset is None:
                # if any unknown shape exist,
                # choose merge method
                return default_tile(op, inp)

            # check subset data to see if it's small enough
            subset_dtypes = inp.dtypes[op.subset]
            memory_usage = 0.
            for s_dtype in subset_dtypes:
                if s_dtype.kind == 'O' or not hasattr(s_dtype, 'itemsize'):
                    # object, just use default tile
                    return default_tile(op, inp)
                else:
                    memory_usage += s_dtype.itemsize * inp.shape[0]
            if memory_usage <= options.chunk_store_limit:
                # if subset is small enough, use method 'subset_tree'
                return cls._tile_subset_tree(op, inp)
            else:
                return default_tile(op, inp)
        elif op.method == 'subset_tree':
            return cls._tile_subset_tree(op, inp)
        elif op.method == 'tree':
            return cls._tile_tree(op, inp)
        else:
            assert op.method == 'shuffle'
            return cls._tile_shuffle(op, inp)

    @classmethod
    def _get_xdf(cls, x):
        if cudf is None:
            return pd
        elif isinstance(x, (pd.Index, pd.Series, pd.DataFrame)):  # pragma: no cover
            return pd
        else:  # pragma: no cover
            return cudf

    @classmethod
    def _drop_duplicates(cls, inp, op, subset=None, keep=None, ignore_index=None):
        if ignore_index is None:
            ignore_index = op.ignore_index
        if subset is None:
            subset = op.subset
        if keep is None:
            keep = op.keep
        if inp.ndim == 2:
            try:
                return inp.drop_duplicates(subset=subset,
                                           keep=keep,
                                           ignore_index=ignore_index)
            except TypeError:
                # no ignore_index for pandas < 1.0
                ret = inp.drop_duplicates(subset=subset,
                                          keep=keep)
                if ignore_index:
                    ret.reset_index(drop=True, inplace=True)
                return ret
        else:
            return inp.drop_duplicates(keep=keep)

    @classmethod
    def _execute_chunk(cls, ctx, op):
        inp = ctx[op.input.key]
        ctx[op.outputs[0].key] = cls._drop_duplicates(inp, op)

    @classmethod
    def _execute_subset_tree_map(cls, ctx, op):
        out = op.outputs[0]
        idx = out.index[0]
        # input contains subset only
        inp = ctx[op.input.key]
        xdf = cls._get_xdf(inp)

        # index would be (chunk_index, i)
        index = xdf.MultiIndex.from_arrays(
            [np.full(inp.shape[0], idx), np.arange(inp.shape[0])],
            names=['_chunk_index_', '_i_'])
        inp = inp.set_index(index)
        ctx[out.key] = cls._drop_duplicates(inp, op, ignore_index=False)

    @classmethod
    def _execute_subset_tree_combine(cls, ctx, op):
        inp = ctx[op.input.key]
        ctx[op.outputs[0].key] = cls._drop_duplicates(inp, op,
                                                      ignore_index=False)

    @classmethod
    def _execute_subset_tree_agg(cls, ctx, op):
        inp = ctx[op.input.key]
        ret = cls._drop_duplicates(inp, op, ignore_index=False)
        ret = ret.index.to_frame()
        ret.reset_index(drop=True, inplace=True)
        ctx[op.outputs[0].key] = ret

    @classmethod
    def _execute_subset_tree_post(cls, ctx, op):
        inp = ctx[op.input.key]
        out = op.outputs[0]
        idx = op.outputs[0].index[0]
        subset = ctx[op.subset_chunk.key]
        selected = subset[subset['_chunk_index_'] == idx]['_i_']
        ret = inp.iloc[selected]
        if op.ignore_index:
            prev_size = (subset['_chunk_index_'] < out.index[0]).sum()
            ret.index = pd.RangeIndex(prev_size, prev_size + len(ret))
        ctx[op.outputs[0].key] = ret

    @classmethod
    def _execute_shuffle_map(cls, ctx, op):
        out = op.outputs[0]
        shuffle_size = op.shuffle_size
        subset = op.subset

        inp = ctx[op.input.key]
        dropped = cls._drop_duplicates(inp, op)
        if dropped.ndim == 1:
            dropped = dropped.to_frame()
            subset = dropped.columns.tolist()
        else:
            if subset is None:
                subset = dropped.columns.tolist()
        dropped['_chunk_index_'] = out.index[0]
        dropped['_i_'] = np.arange(dropped.shape[0])
        hashed = hash_dataframe_on(dropped, subset, shuffle_size)
        for i, data in enumerate(hashed):
            ctx[(out.key, str(i))] = dropped.loc[data]

    @classmethod
    def _execute_shuffle_reduce(cls, ctx, op):
        out = op.outputs[0]

        input_keys, _ = get_shuffle_input_keys_idxes(op.inputs[0])
        inputs = [ctx[(inp_key, str(out.index[0]))] for inp_key in input_keys]
        xdf = cls._get_xdf(inputs[0])
        inp = xdf.concat(inputs)
        dropped = cls._drop_duplicates(inp, op,
                                       subset=[c for c in inp.columns
                                               if c not in ('_chunk_index_', '_i_')],
                                       keep=op.keep, ignore_index=op.ignore_index)
        for i in range(op.shuffle_size):
            filtered = dropped[dropped['_chunk_index_'] == i]
            del filtered['_chunk_index_']
            ctx[(out.key, str(i))] = filtered

    @classmethod
    def _execute_shuffle_put_back(cls, ctx, op):
        out = op.outputs[0]

        input_keys, _ = get_shuffle_input_keys_idxes(op.inputs[0])
        inputs = [ctx[(inp_key, str(out.index[0]))] for inp_key in input_keys]
        xdf = cls._get_xdf(inputs[0])
        inp = xdf.concat(inputs)
        inp.sort_values('_i_', inplace=True)
        del inp['_i_']

        if out.op.output_types[0] == OutputType.index:
            assert inp.shape[1] == 1
            ret = xdf.Index(inp.iloc[:, 0])
        elif out.op.output_types[0] == OutputType.series:
            assert inp.shape[1] == 1
            ret = inp.iloc[:, 0]
        else:
            ret = inp

        if op.ignore_index:
            ret.reset_index(drop=True, inplace=True)
        ctx[op.outputs[0].key] = ret

    @classmethod
    def execute(cls, ctx, op):
        if op.method is None:
            # one chunk
            cls._execute_chunk(ctx, op)
        elif op.method == 'tree':
            # tree
            cls._execute_chunk(ctx, op)
        elif op.method == 'subset_tree':
            # subset tree
            if op.stage == OperandStage.map:
                cls._execute_subset_tree_map(ctx, op)
            elif op.stage == OperandStage.combine:
                cls._execute_subset_tree_combine(ctx, op)
            elif op.stage == OperandStage.agg:
                cls._execute_subset_tree_agg(ctx, op)
            else:
                # post
                cls._execute_subset_tree_post(ctx, op)
        else:
            assert op.method == 'shuffle'
            if op.stage == OperandStage.map:
                cls._execute_shuffle_map(ctx, op)
            elif op.shuffle_phase == 'drop_duplicates':
                cls._execute_shuffle_reduce(ctx, op)
            else:
                assert op.shuffle_phase == 'put_back'
                cls._execute_shuffle_put_back(ctx, op)


def _validate_subset(df, subset):
    if subset is None:
        return subset
    if not is_list_like(subset):
        subset = [subset]
    else:
        subset = list(subset)

    for s in subset:
        if s not in df.dtypes:
            raise KeyError(pd.Index([s]))

    return subset


def df_drop_duplicates(df, subset=None, keep='first',
                       inplace=False, ignore_index=False, method='auto'):
    """
    Return DataFrame with duplicate rows removed.

    Considering certain columns is optional. Indexes, including time indexes
    are ignored.

    Parameters
    ----------
    subset : column label or sequence of labels, optional
        Only consider certain columns for identifying duplicates, by
        default use all of the columns.
    keep : {'first', 'last', False}, default 'first'
        Determines which duplicates (if any) to keep.
        - ``first`` : Drop duplicates except for the first occurrence.
        - ``last`` : Drop duplicates except for the last occurrence.
        - False : Drop all duplicates.
    inplace : bool, default False
        Whether to drop duplicates in place or to return a copy.
    ignore_index : bool, default False
        If True, the resulting axis will be labeled 0, 1, …, n - 1.

    Returns
    -------
    DataFrame
        DataFrame with duplicates removed or None if ``inplace=True``.
    """
    if method not in ('auto', 'tree', 'subset_tree', 'shuffle', None):
        raise ValueError("method could only be one of "
                         "'auto', 'tree', 'subset_tree', 'shuffle' or None")
    subset = _validate_subset(df, subset)
    op = DataFrameDropDuplicates(subset=subset, keep=keep,
                                 ignore_index=ignore_index,
                                 method=method)
    return op(df, inplace=inplace)


def series_drop_duplicates(series, keep='first', inplace=False, method='auto'):
    """
    Return Series with duplicate values removed.

    Parameters
    ----------
    keep : {'first', 'last', ``False``}, default 'first'
        Method to handle dropping duplicates:

        - 'first' : Drop duplicates except for the first occurrence.
        - 'last' : Drop duplicates except for the last occurrence.
        - ``False`` : Drop all duplicates.

    inplace : bool, default ``False``
        If ``True``, performs operation inplace and returns None.

    Returns
    -------
    Series
        Series with duplicates dropped.

    See Also
    --------
    Index.drop_duplicates : Equivalent method on Index.
    DataFrame.drop_duplicates : Equivalent method on DataFrame.
    Series.duplicated : Related method on Series, indicating duplicate
        Series values.

    Examples
    --------
    Generate a Series with duplicated entries.

    >>> import mars.dataframe as md
    >>> s = md.Series(['lama', 'cow', 'lama', 'beetle', 'lama', 'hippo'],
    ...               name='animal')
    >>> s.execute()
    0      lama
    1       cow
    2      lama
    3    beetle
    4      lama
    5     hippo
    Name: animal, dtype: object

    With the 'keep' parameter, the selection behaviour of duplicated values
    can be changed. The value 'first' keeps the first occurrence for each
    set of duplicated entries. The default value of keep is 'first'.

    >>> s.drop_duplicates().execute()
    0      lama
    1       cow
    3    beetle
    5     hippo
    Name: animal, dtype: object

    The value 'last' for parameter 'keep' keeps the last occurrence for
    each set of duplicated entries.

    >>> s.drop_duplicates(keep='last').execute()
    1       cow
    3    beetle
    4      lama
    5     hippo
    Name: animal, dtype: object

    The value ``False`` for parameter 'keep' discards all sets of
    duplicated entries. Setting the value of 'inplace' to ``True`` performs
    the operation inplace and returns ``None``.

    >>> s.drop_duplicates(keep=False, inplace=True)
    >>> s.execute()
    1       cow
    3    beetle
    5     hippo
    Name: animal, dtype: object
    """
    if method not in ('auto', 'tree', 'shuffle', None):
        raise ValueError("method could only be one of "
                         "'auto', 'tree', 'shuffle' or None")
    op = DataFrameDropDuplicates(keep=keep, method=method)
    return op(series, inplace=inplace)


def index_drop_duplicates(index, keep='first', method='auto'):
    """
    Return Index with duplicate values removed.

    Parameters
    ----------
    keep : {'first', 'last', ``False``}, default 'first'
        - 'first' : Drop duplicates except for the first occurrence.
        - 'last' : Drop duplicates except for the last occurrence.
        - ``False`` : Drop all duplicates.

    Returns
    -------
    deduplicated : Index

    See Also
    --------
    Series.drop_duplicates : Equivalent method on Series.
    DataFrame.drop_duplicates : Equivalent method on DataFrame.
    Index.duplicated : Related method on Index, indicating duplicate
        Index values.

    Examples
    --------
    Generate an pandas.Index with duplicate values.

    >>> import mars.dataframe as md

    >>> idx = md.Index(['lama', 'cow', 'lama', 'beetle', 'lama', 'hippo'])

    The `keep` parameter controls  which duplicate values are removed.
    The value 'first' keeps the first occurrence for each
    set of duplicated entries. The default value of keep is 'first'.

    >>> idx.drop_duplicates(keep='first').execute()
    Index(['lama', 'cow', 'beetle', 'hippo'], dtype='object')

    The value 'last' keeps the last occurrence for each set of duplicated
    entries.

    >>> idx.drop_duplicates(keep='last').execute()
    Index(['cow', 'beetle', 'lama', 'hippo'], dtype='object')

    The value ``False`` discards all sets of duplicated entries.

    >>> idx.drop_duplicates(keep=False).execute()
    Index(['cow', 'beetle', 'hippo'], dtype='object')
    """
    if method not in ('auto', 'tree', 'shuffle', None):
        raise ValueError("method could only be one of "
                         "'auto', 'tree', 'shuffle' or None")
    op = DataFrameDropDuplicates(keep=keep, method=method)
    return op(index)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict

from ....serialize import AnyField, Int64Field, BoolField, StringField, Int32Field
from ...core import DATAFRAME_TYPE
from ...utils import build_empty_df, build_empty_series, validate_axis
from ..core import Window


class Rolling(Window):
    _window = AnyField('window')
    _min_periods = Int64Field('min_periods')
    _center = BoolField('center')
    _win_type = StringField('win_type')
    _on = StringField('on')
    _axis = Int32Field('axis')
    _closed = StringField('closed')

    def __init__(self, window=None, min_periods=None, center=None, win_type=None, on=None,
                 axis=None, closed=None, **kw):
        super().__init__(_window=window, _min_periods=min_periods, _center=center,
                         _win_type=win_type, _on=on, _axis=axis, _closed=closed, **kw)

    @property
    def window(self):
        return self._window

    @property
    def min_periods(self):
        return self._min_periods

    @property
    def center(self):
        return self._center

    @property
    def win_type(self):
        return self._win_type

    @property
    def on(self):
        return self._on

    @property
    def axis(self):
        return self._axis

    @property
    def closed(self):
        return self._closed

    @property
    def params(self):
        p = OrderedDict()
        for attr in ['window', 'min_periods', 'center',
                     'win_type', 'axis', 'on', 'closed']:
            p[attr] = getattr(self, attr)
        return p

    def _repr_name(self):
        return 'Rolling' if self.win_type is None else 'Window'

    def validate(self):
        # leverage pandas itself to do validation
        pd_index = self._input.index_value.to_pandas()
        if isinstance(self._input, DATAFRAME_TYPE):
            empty_obj = build_empty_df(self._input.dtypes, index=pd_index[:0])
        else:
            empty_obj = build_empty_series(self._input.dtype, index=pd_index[:0],
                                           name=self._input.name)
        pd_rolling = empty_obj.rolling(**self.params)
        for k in self.params:
            # update value according to pandas rolling
            setattr(self, '_' + k, getattr(pd_rolling, k))

    def aggregate(self, func, *args, **kwargs):
        from .aggregation import DataFrameRollingAgg

        op = DataFrameRollingAgg(func=func, func_args=args,
                                 func_kwargs=kwargs, **self.params)
        return op(self)

    def agg(self, func, *args, **kwargs):
        return self.aggregate(func, *args, **kwargs)

    def count(self):
        return self.aggregate('count')

    def sum(self, *args, **kwargs):
        return self.aggregate('sum', *args, **kwargs)

    def mean(self, *args, **kwargs):
        return self.aggregate('mean', *args, **kwargs)

    def median(self, **kwargs):
        return self.aggregate('median', **kwargs)

    def var(self, ddof=1, *args, **kwargs):
        return self.aggregate('var', ddof=ddof, *args, **kwargs)

    def std(self, ddof=1, *args, **kwargs):
        return self.aggregate('std', ddof=ddof, *args, **kwargs)

    def min(self, *args, **kwargs):
        return self.aggregate('min', *args, **kwargs)

    def max(self, *args, **kwargs):
        return self.aggregate('max', *args, **kwargs)

    def skew(self, **kwargs):
        return self.aggregate('skew', **kwargs)

    def kurt(self, **kwargs):
        return self.aggregate('kurt', **kwargs)


def rolling(obj, window, min_periods=None, center=False, win_type=None, on=None,
            axis=0, closed=None):
    """
    Provide rolling window calculations.

    Parameters
    ----------
    window : int, or offset
        Size of the moving window. This is the number of observations used for
        calculating the statistic. Each window will be a fixed size.
        If its an offset then this will be the time period of each window. Each
        window will be a variable sized based on the observations included in
        the time-period. This is only valid for datetimelike indexes. This is
        new in 0.19.0
    min_periods : int, default None
        Minimum number of observations in window required to have a value
        (otherwise result is NA). For a window that is specified by an offset,
        `min_periods` will default to 1. Otherwise, `min_periods` will default
        to the size of the window.
    center : bool, default False
        Set the labels at the center of the window.
    win_type : str, default None
        Provide a window type. If ``None``, all points are evenly weighted.
        See the notes below for further information.
    on : str, optional
        For a DataFrame, a datetime-like column on which to calculate the rolling
        window, rather than the DataFrame's index. Provided integer column is
        ignored and excluded from result since an integer index is not used to
        calculate the rolling window.
    axis : int or str, default 0
    closed : str, default None
        Make the interval closed on the 'right', 'left', 'both' or
        'neither' endpoints.
        For offset-based windows, it defaults to 'right'.
        For fixed windows, defaults to 'both'. Remaining cases not implemented
        for fixed windows.

    Returns
    -------
    a Window or Rolling sub-classed for the particular operation

    See Also
    --------
    expanding : Provides expanding transformations.
    ewm : Provides exponential weighted functions.

    Notes
    -----
    By default, the result is set to the right edge of the window. This can be
    changed to the center of the window by setting ``center=True``.
    To learn more about the offsets & frequency strings, please see `this link
    <http://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases>`__.

    The recognized win_types are:
    * ``boxcar``
    * ``triang``
    * ``blackman``
    * ``hamming``
    * ``bartlett``
    * ``parzen``
    * ``bohman``
    * ``blackmanharris``
    * ``nuttall``
    * ``barthann``
    * ``kaiser`` (needs beta)
    * ``gaussian`` (needs std)
    * ``general_gaussian`` (needs power, width)
    * ``slepian`` (needs width)
    * ``exponential`` (needs tau), center is set to None.

    If ``win_type=None`` all points are evenly weighted. To learn more about
    different window types see `scipy.signal window functions
    <https://docs.scipy.org/doc/scipy/reference/signal.html#window-functions>`__.

    Examples
    --------
    >>> import numpy as np
    >>> import mars.dataframe as md
    >>> df = md.DataFrame({'B': [0, 1, 2, np.nan, 4]})
    >>> df.execute()
         B
    0  0.0
    1  1.0
    2  2.0
    3  NaN
    4  4.0

    Rolling sum with a window length of 2, using the 'triang'
    window type.

    >>> df.rolling(2, win_type='triang').sum().execute()
         B
    0  NaN
    1  0.5
    2  1.5
    3  NaN
    4  NaN

    Rolling sum with a window length of 2, min_periods defaults
    to the window length.

    >>> df.rolling(2).sum().execute()
         B
    0  NaN
    1  1.0
    2  3.0
    3  NaN
    4  NaN

    Same as above, but explicitly set the min_periods

    >>> df.rolling(2, min_periods=1).sum().execute()
         B
    0  0.0
    1  1.0
    2  3.0
    3  2.0
    4  4.0

    A ragged (meaning not-a-regular frequency), time-indexed DataFrame

    >>> df = md.DataFrame({'B': [0, 1, 2, np.nan, 4]},
    >>>                   index = [md.Timestamp('20130101 09:00:00'),
    >>>                            md.Timestamp('20130101 09:00:02'),
    >>>                            md.Timestamp('20130101 09:00:03'),
    >>>                            md.Timestamp('20130101 09:00:05'),
    >>>                            md.Timestamp('20130101 09:00:06')])
    >>> df.execute()
                           B
    2013-01-01 09:00:00  0.0
    2013-01-01 09:00:02  1.0
    2013-01-01 09:00:03  2.0
    2013-01-01 09:00:05  NaN
    2013-01-01 09:00:06  4.0

    Contrasting to an integer rolling window, this will roll a variable
    length window corresponding to the time period.
    The default for min_periods is 1.

    >>> df.rolling('2s').sum().execute()
                           B
    2013-01-01 09:00:00  0.0
    2013-01-01 09:00:02  1.0
    2013-01-01 09:00:03  3.0
    2013-01-01 09:00:05  NaN
    2013-01-01 09:00:06  4.0
    """
    axis = validate_axis(axis, obj)
    r = Rolling(input=obj, window=window, min_periods=min_periods, center=center,
                win_type=win_type, on=on, axis=axis, closed=closed)
    r.validate()
    return r



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import pickle
import sys
import weakref
from collections import OrderedDict
from datetime import tzinfo
from io import BytesIO

import cloudpickle
import numpy as np
cimport numpy as np
import pandas as pd
from pandas.api.extensions import ExtensionDtype
from pandas.arrays import IntervalArray

from ..utils import serialize_function
from .core cimport Provider, ValueType, ProviderType, \
    Field, List, Tuple, Dict, Identity, Reference, KeyPlaceholder, \
    ReferenceField, OneOfField, ListField, get_serializable_by_index
from .core import HasKey, HasData
from .dataserializer import dumps as datadumps, loads as dataloads

try:
    from pandas.tseries.offsets import Tick as PDTick
    from pandas.tseries.frequencies import to_offset
except ImportError:
    PDTick = to_offset = None


cdef dict PRIMITIVE_TYPE_TO_NAME = {
    ValueType.bool: 'bool',
    ValueType.int8: 'int8',
    ValueType.int16: 'int16',
    ValueType.int32: 'int32',
    ValueType.int64: 'int64',
    ValueType.uint8: 'uint8',
    ValueType.uint16: 'uint16',
    ValueType.uint32: 'uint32',
    ValueType.uint64: 'uint64',
    ValueType.float16: 'float16',
    ValueType.float32: 'float32',
    ValueType.float64: 'float64',
    ValueType.bytes: 'bytes',
    ValueType.unicode: 'unicode',
}

cdef dict EXTEND_TYPE_TO_NAME = {
    ValueType.slice: 'slice',
    ValueType.arr: 'arr',
    ValueType.dtype: 'dtype',
    ValueType.index: 'index',
    ValueType.series: 'series',
    ValueType.dataframe: 'dataframe',
    ValueType.key: 'key',
    ValueType.datetime64: 'datetime64',
    ValueType.timedelta64: 'timedelta64',
    ValueType.complex64: 'complex64',
    ValueType.complex128: 'complex128',
}


cdef inline str _get_name(value_type):
    if value_type in PRIMITIVE_TYPE_TO_NAME:
        return PRIMITIVE_TYPE_TO_NAME[value_type]
    if value_type in EXTEND_TYPE_TO_NAME:
        return EXTEND_TYPE_TO_NAME[value_type]
    return value_type.name


cdef inline object _get_type(str value_type_name):
    return getattr(ValueType, value_type_name)


cdef class JsonSerializeProvider(Provider):
    def __init__(self, data_serial_type=None, pickle_protocol=None):
        self.type = ProviderType.json
        self.data_serial_type = data_serial_type
        self.pickle_protocol = pickle_protocol

    cdef inline str _to_str(self, val):
        assert isinstance(val, (bytes, unicode))
        if unicode is str and isinstance(val, bytes):
            return val.decode('utf-8')
        return val

    cdef inline bytes _to_bytes(self, val):
        assert isinstance(val, (bytes, unicode))
        if isinstance(val, unicode):
            return val.encode('utf-8')
        return val

    cdef inline dict _serialize_start_stop(self, pos):
        if isinstance(pos, int):
            return {'type': 'int',
                    'value': pos}
        elif isinstance(pos, str):
            return {'type': 'str',
                    'value': pos}
        else:
            return {'type': 'object',
                    'value': self._to_str(base64.b64encode(
                        pickle.dumps(pos, protocol=self.pickle_protocol)))}

    cdef inline dict _serialize_slice(self, slice value):
        return {
            'type': _get_name(ValueType.slice),
            'value': {
                'start': self._serialize_start_stop(value.start),
                'stop': self._serialize_start_stop(value.stop),
                'step': value.step
            }
        }

    cdef inline object _deserialize_start_stop(self, value):
        if value['type'] in ('int', 'str'):
            return value['value']
        else:
            return pickle.loads(base64.b64decode(value['value']))

    cdef inline slice _deserialize_slice(self, object obj, list callbacks):
        value = obj['value']
        return slice(self._deserialize_start_stop(value['start']),
                     self._deserialize_start_stop(value['stop']),
                     value['step'])

    cdef inline dict _serialize_arr(self, np.ndarray value):
        # special case for np.unicode and np.bytes_
        # cuz datadumps may fail due to pyarrow
        if value.ndim == 0 and value.dtype.kind in ('U', 'S'):
            value = value.astype(object)
        return {
            'type': _get_name(ValueType.arr),
            'value': self._to_str(base64.b64encode(datadumps(value, serial_type=self.data_serial_type,
                                                             pickle_protocol=self.pickle_protocol)))
        }

    cdef inline np.ndarray _deserialize_arr(self, object obj, list callbacks):
        cdef bytes bt

        value = obj['value']

        bt = self._to_bytes(base64.b64decode(value))

        if bt is not None:
            return dataloads(bt)

        return None

    cdef inline dict _serialize_dtype(self, value):
        cdef str v

        if not isinstance(value, ExtensionDtype) and 'V' not in value.str:
            v = value.str
        else:
            v = self._to_str(base64.b64encode(pickle.dumps(value, protocol=self.pickle_protocol)))
        return {
            'type': _get_name(ValueType.dtype),
            'value': v
        }


    cdef inline object _deserialize_dtype(self, object value, list callbacks):
        try:
            return np.dtype(value['value'])
        except TypeError:
            val = value['value']
            return pickle.loads(base64.b64decode(val))

    cdef inline dict _serialize_index(self, value):
        return {
            'type': _get_name(ValueType.index),
            'value': self._to_str(base64.b64encode(datadumps(value, serial_type=self.data_serial_type,
                                                             pickle_protocol=self.pickle_protocol)))
        }

    cdef inline object _deserialize_pd_entity(self, object obj, list callbacks):
        cdef bytes bt

        value = obj['value']

        bt = self._to_bytes(base64.b64decode(value))

        if bt is not None:
            return dataloads(bt)

        return None

    cdef inline dict _serialize_series(self, value):
        return {
            'type': _get_name(ValueType.series),
            'value': self._to_str(base64.b64encode(datadumps(value, serial_type=self.data_serial_type,
                                                             pickle_protocol=self.pickle_protocol)))
        }

    cdef inline dict _serialize_dataframe(self, value):
        return {
            'type': _get_name(ValueType.dataframe),
            'value': self._to_str(base64.b64encode(datadumps(value, serial_type=self.data_serial_type,
                                                             pickle_protocol=self.pickle_protocol)))
        }

    cdef inline dict _serialize_key(self, value):
        return {
            'type': _get_name(ValueType.key),
            'value': (value.key, value.id)
        }

    cdef inline KeyPlaceholder _deserialize_key(self, object obj, list callbacks):
        return KeyPlaceholder(*obj['value'])

    cdef inline dict _serialize_list(self, list value, tp=None, bint weak_ref=False):
        return {
            'type': _get_name(ValueType.list),
            'value': [self._serialize_value(val if not weak_ref else val(),
                                            tp.type if tp is not None else None)
                      for val in value]
        }

    cdef inline list _deserialize_list(self, object obj, list callbacks, bint weak_ref):
        cdef list res
        cdef int i
        cdef int j

        res = []
        for i, it_val in enumerate(obj['value']):
            val = self._deserialize_value(it_val, callbacks, weak_ref)
            res.append(val)
            if isinstance(val, KeyPlaceholder):
                def cb(j, v):
                    def inner(subs):
                        o = subs[v.key, v.id]
                        if weak_ref:
                            o = weakref.ref(o)
                        res[j] = o
                    return inner
                callbacks.append(cb(i, val))

        return res

    cdef inline dict _serialize_tuple(self, tuple value, tp=None, bint weak_ref=False):
        if tp is None or not isinstance(tp.type, tuple):
            return {
                'type': _get_name(ValueType.tuple),
                'value': [self._serialize_value(val if not weak_ref else val(),
                                                tp.type if tp is not None else None)
                          for val in value]
            }
        else:
            return {
                'type': _get_name(ValueType.tuple),
                'value': [self._serialize_value(val if not weak_ref else val(),
                                                it_type)
                          for val, it_type in zip(value, tp.type)]
            }

    cdef inline tuple _deserialize_tuple(self, object obj, list callbacks, bint weak_ref):
        return tuple(self._deserialize_value(it_val, callbacks, weak_ref)
                     for it_val in obj['value'])

    cdef inline dict _serialize_dict(self, object value, tp=None, bint weak_ref=False):
        return {
            'type': _get_name(ValueType.dict),
            'value': [(self._serialize_value(k if not weak_ref else k(),
                                             tp.key_type if tp is not None else tp),
                       self._serialize_value(v if not weak_ref else v(),
                                             tp.value_type if tp is not None else tp))
                      for k, v in value.items()]
        }

    cdef inline object _deserialize_dict(self, object obj, list callbacks, bint weak_ref):
        res = OrderedDict()
        for k, v in obj['value']:
            key = self._deserialize_value(k, callbacks, weak_ref)
            value = self._deserialize_value(v, callbacks, weak_ref)
            res[key] = value
            if isinstance(value, KeyPlaceholder):
                def cb(ko, vo):
                    def inner(subs):
                        o = subs[vo.key, vo.id]
                        if weak_ref:
                            o = weakref.ref(o)
                        res[ko] = o
                    return inner
                callbacks.append(cb(key, value))

        return res

    cdef inline dict _serialize_datetime64_timedelta64(self, value, tp):
        bio = BytesIO()
        np.save(bio, value)
        return {
            'type': _get_name(tp),
            'value': self._to_str(base64.b64encode(bio.getvalue()))
        }

    cdef inline _deserialize_datetime64_timedelta64(self, obj, list callbacks):
        cdef bytes v

        value = obj['value']
        v = base64.b64decode(value)

        if v is not None:
            # np.load will return a ndarray
            value = np.load(BytesIO(v))
            return value.dtype.type(value)
        return None

    cdef inline dict _serialize_datetime64(self, value):
        if isinstance(value, pd.Timestamp):
            # convert to np.datetime64
            value = value.to_datetime64()
        return self._serialize_datetime64_timedelta64(value, ValueType.datetime64)

    cdef inline _deserialize_datetime64(self, obj, list callbacks):
        return self._deserialize_datetime64_timedelta64(obj, callbacks)

    cdef inline dict _serialize_timedelta64(self, value):
        if isinstance(value, pd.Timedelta):
            # convert to np.timedelta64
            value = value.to_timedelta64()
        return self._serialize_datetime64_timedelta64(value, ValueType.timedelta64)

    cdef inline _deserialize_timedelta64(self, obj, list callbacks):
        return self._deserialize_datetime64_timedelta64(obj, callbacks)

    cdef inline dict _serialize_complex(self, value, tp):
        return {
            'type': _get_name(tp),
            'value': (value.real, value.imag)
        }

    cdef inline _deserialize_complex(self, obj, list callbacks):
        cdef list v

        v = obj['value']
        return complex(*v)

    cdef inline dict _serialize_function(self, value):
        return {
            'type': 'function',
            'value': self._to_str(base64.b64encode(
                serialize_function(value, pickle_protocol=self.pickle_protocol)))
        }

    cdef inline _deserialize_function(self, obj, list callbacks):
        cdef bytes v

        value = obj['value']
        v = base64.b64decode(value)

        if v is not None:
            return cloudpickle.loads(v)
        return None

    cdef inline dict _serialize_tzinfo(self, value):
        return {
            'type': 'tzinfo',
            'value': self._to_str(base64.b64encode(pickle.dumps(value, protocol=self.pickle_protocol)))
        }

    cdef inline _deserialize_tzinfo(self, obj, list callbacks):
        cdef bytes v

        value = obj['value']
        v = base64.b64decode(value)

        if v is not None:
            return pickle.loads(v)
        return None

    cdef inline _serialize_interval_arr(self, value):
        return {
            'type': 'interval_arr',
            'value': {
                'left': self._to_str(base64.b64encode(datadumps(value.left, serial_type=self.data_serial_type,
                                                                pickle_protocol=self.pickle_protocol))),
                'right': self._to_str(base64.b64encode(datadumps(value.right, serial_type=self.data_serial_type,
                                                                 pickle_protocol=self.pickle_protocol))),
                'closed': value.closed,
                'dtype': self._serialize_dtype(value.dtype)
            }
        }

    cdef inline _deserialize_interval_arr(self, obj, list callbacks):
        value = obj['value']

        left = dataloads(self._to_bytes(base64.b64decode(value['left'])))
        right = dataloads(self._to_bytes(base64.b64decode(value['right'])))
        closed = value['closed']
        dtype = self._deserialize_dtype(value['dtype'], callbacks)

        return IntervalArray.from_arrays(left, right, closed=closed, dtype=dtype)

    cdef inline _serialize_freq(self, value):
        return {
            'type': 'freq',
            'value': value.freqstr,
        }

    cdef inline _deserialize_freq(self, obj, list callbacks):
        value = obj['value']
        return to_offset(value)

    cdef inline object _serialize_typed_value(self, value, tp, bint weak_ref=False):
        if type(tp) not in (List, Tuple, Dict) and weak_ref:
            # not iterable, and is weak ref
            value = value()
        if value is None:
            return
        if tp == ValueType.bytes:
            return {
                'type': _get_name(tp),
                'value': self._to_str(base64.b64encode(value))
            }
        elif tp in PRIMITIVE_TYPE_TO_NAME:
            # primitive type, we do not do any type check here
            return value
        elif type(tp) is Identity:
            return self._serialize_typed_value(value, tp.type, weak_ref=weak_ref)
        elif tp in {ValueType.complex64, ValueType.complex128}:
            return self._serialize_complex(value, tp)
        elif tp is ValueType.slice:
            return self._serialize_slice(value)
        elif tp is ValueType.arr:
            return self._serialize_arr(value)
        elif tp is ValueType.dtype:
            return self._serialize_dtype(value)
        elif tp is ValueType.index:
            return self._serialize_index(value)
        elif tp is ValueType.series:
            return self._serialize_series(value)
        elif tp is ValueType.dataframe:
            return self._serialize_dataframe(value)
        elif tp is ValueType.key:
            return self._serialize_key(value)
        elif tp == ValueType.datetime64:
            return self._serialize_datetime64(value)
        elif tp == ValueType.timedelta64:
            return self._serialize_timedelta64(value)
        elif tp == ValueType.function:
            return self._serialize_function(value)
        elif tp == ValueType.tzinfo:
            return self._serialize_tzinfo(value)
        elif tp == ValueType.interval_arr:
            return self._serialize_interval_arr(value)
        elif tp == ValueType.freq:
            return self._serialize_freq(value)
        elif isinstance(tp, List):
            if not isinstance(value, list):
                value = list(value)
            return self._serialize_list(value, tp, weak_ref=weak_ref)
        elif isinstance(tp, Tuple):
            return self._serialize_tuple(value, tp, weak_ref=weak_ref)
        elif isinstance(tp, Dict):
            return self._serialize_dict(value, tp, weak_ref=weak_ref)
        else:
            raise TypeError(f'Unknown type to serialize: {tp}')

    cdef inline object _serialize_untyped_value(self, value, bint weak_ref=False):
        if not isinstance(value, (list, tuple, dict)) and weak_ref:
            # not iterable, and is weak ref
            value = value()
        if value is None:
            return

        if isinstance(value, bool):
            return value
        elif isinstance(value, bytes):
            return {
                'type': _get_name(ValueType.bytes),
                'value': self._to_str(base64.b64encode(value)),
            }
        elif isinstance(value, unicode):
            return value
        elif isinstance(value, int):
            return value
        elif isinstance(value, float):
            return value
        elif isinstance(value, complex):
            return self._serialize_complex(value, ValueType.complex128)
        elif isinstance(value, slice):
            return self._serialize_slice(value)
        elif isinstance(value, np.ndarray):
            return self._serialize_arr(value)
        elif isinstance(value, (np.dtype, ExtensionDtype)):
            return self._serialize_dtype(value)
        elif isinstance(value, pd.Index):
            return self._serialize_index(value)
        elif isinstance(value, pd.Series):
            return self._serialize_series(value)
        elif isinstance(value, pd.DataFrame):
            return self._serialize_dataframe(value)
        elif isinstance(value, HasKey):
            return self._serialize_key(value)
        elif isinstance(value, HasData):
            return self._serialize_key(value.data)
        elif isinstance(value, list):
            return self._serialize_list(value, tp=None, weak_ref=weak_ref)
        elif isinstance(value, tuple):
            return self._serialize_tuple(value, tp=None, weak_ref=weak_ref)
        elif isinstance(value, dict):
            return self._serialize_dict(value, tp=None, weak_ref=weak_ref)
        elif isinstance(value, (np.datetime64, pd.Timestamp)):
            return self._serialize_datetime64(value)
        elif isinstance(value, (np.timedelta64, pd.Timedelta)):
            return self._serialize_timedelta64(value)
        elif isinstance(value, np.number):
            return self._serialize_untyped_value(value.item())
        elif isinstance(value, tzinfo):
            return self._serialize_tzinfo(value)
        elif isinstance(value, IntervalArray):
            return self._serialize_interval_arr(value)
        elif PDTick is not None and isinstance(value, PDTick):
            return self._serialize_freq(value)
        elif callable(value):
            return self._serialize_function(value)
        else:
            raise TypeError(f'Unknown type to serialize: {type(value)}')

    cdef inline object _serialize_value(self, value, tp=None, bint weak_ref=False):
        if tp is None:
            return self._serialize_untyped_value(value, weak_ref=weak_ref)
        else:
            return self._serialize_typed_value(value, tp, weak_ref=weak_ref)

    cdef inline _on_serial(self, Field field, x):
        x = x if x is not None else field.default
        return field.on_serialize(x) if field.on_serialize is not None else x

    cdef object _serialize_reference(self, tag, model, value, new_obj):
        if value is None:
            return None
        if model is None:
            new_obj['type_id'] = value.__serializable_index__
            new_obj['value'] = value.serialize(self, dict())
        else:
            if not isinstance(value, model):
                raise TypeError(f'Does not match type for reference field {tag}: '
                                f'expect {model}, got {type(value)}')
            value.serialize(self, new_obj)
        return new_obj

    cpdef serialize_field(self, Field field, model_instance, obj):
        cdef str tag
        cdef object new_obj
        cdef bint has_val
        cdef object value
        cdef OneOfField oneoffield

        if isinstance(field, ReferenceField):
            tag = field.tag_name(self)
            if hasattr(model_instance, field.attr):
                field_val = getattr(model_instance, field.attr)
                if field.weak_ref:
                    field_val = field_val()
                if field_val is not None:
                    new_obj = obj[tag] = dict()
                    value = self._on_serial(field, field_val)
                    self._serialize_reference(tag, field.type.model, value, new_obj)
        elif isinstance(field, OneOfField):
            has_val = False
            field_val = getattr(model_instance, field.attr, None)
            if field.weak_ref and field_val is not None:
                field_val = field_val()
            value = self._on_serial(field, field_val)
            oneoffield = <OneOfField>field
            for f in oneoffield.fields:
                tag = f.tag_name(self)
                if isinstance(value, f.type.model):
                    if not has_val:
                        has_val = True
                    else:
                        raise ValueError(f'Only one of attributes({field.attrs}) can be specified')

                    new_obj = obj[tag] = dict()
                    value.serialize(self, new_obj)
                    return
            if not has_val and value is not None:
                raise ValueError(f'Value {value} cannot match any type for OneOfField `{field.tag_name(self)}`')
        elif isinstance(field, ListField) and type(field.type.type) == Reference:
            tag = field.tag_name(self)
            value = self._on_serial(field, getattr(model_instance, field.attr, None))
            if value is None:
                return
            new_obj = obj[tag] = list()
            for val in value:
                if field.weak_ref:
                    val = val()
                new_obj.append(self._serialize_reference(tag, field.type.type.model, val, dict()))
        else:
            tag = field.tag_name(self)
            try:
                val = self._on_serial(field, getattr(model_instance, field.attr, None))
            except (AttributeError, TypeError):
                tp, err, tb = sys.exc_info()
                raise tp(f'Fail to serialize field `{tag}` for {model_instance}, reason: {err}') \
                    .with_traceback(tb) from err
            if val is None:
                return
            try:
                obj[tag] = self._serialize_value(val, field.type, weak_ref=field.weak_ref)
            except (TypeError, ValueError):
                tp, err, tb = sys.exc_info()
                raise tp(f'Fail to serialize field `{tag}` for {model_instance}, reason: {err}') \
                    .with_traceback(tb) from err

    cdef inline _deserialize_value(self, obj, list callbacks, bint weak_ref):
        if not isinstance(obj, dict):
            return obj

        ref = lambda x: weakref.ref(x) if weak_ref else x
        tp = _get_type(obj['type'])

        if tp is ValueType.bytes:
            return ref(base64.b64decode(obj['value']))
        elif tp in {ValueType.complex64, ValueType.complex128}:
            return ref(self._deserialize_complex(obj, callbacks))
        elif tp is ValueType.slice:
            return ref(self._deserialize_slice(obj, callbacks))
        elif tp is ValueType.arr:
            return ref(self._deserialize_arr(obj, callbacks))
        elif tp is ValueType.dtype:
            return ref(self._deserialize_dtype(obj, callbacks))
        elif tp in (ValueType.index, ValueType.series, ValueType.dataframe):
            return ref(self._deserialize_pd_entity(obj, callbacks))
        elif tp is ValueType.key:
            return self._deserialize_key(obj, callbacks)  # the weakref will do in the callback, so skip
        elif tp is ValueType.datetime64:
            return ref(self._deserialize_datetime64(obj, callbacks))
        elif tp is ValueType.timedelta64:
            return ref(self._deserialize_timedelta64(obj, callbacks))
        elif tp is ValueType.function:
            return ref(self._deserialize_function(obj, callbacks))
        elif tp is ValueType.tzinfo:
            return ref(self._deserialize_tzinfo(obj, callbacks))
        elif tp is ValueType.interval_arr:
            return ref(self._deserialize_interval_arr(obj, callbacks))
        elif tp is ValueType.freq:
            return ref(self._deserialize_freq(obj, callbacks))
        elif tp is ValueType.list:
            return self._deserialize_list(obj, callbacks, weak_ref)
        elif tp is ValueType.tuple:
            return self._deserialize_tuple(obj, callbacks, weak_ref)
        elif tp is ValueType.dict:
            return self._deserialize_dict(obj, callbacks, weak_ref)
        else:
            raise TypeError(f'Unknown type to deserialize {obj["type"]}')

    cdef inline _on_deserial(self, Field field, x):
        x = x if x is not None else field.default
        return field.on_deserialize(x) if field.on_deserialize is not None else x

    cdef object _deserialize_reference(self, model, val, list callbacks, dict key_to_instance):
        if model is None:
            model = get_serializable_by_index(val['type_id'])
            if model is None:
                raise KeyError(f'Cannot find serializable class for type_id {val["type_id"]}')
            val = val['value']
        return model.deserialize(self, val, callbacks, key_to_instance)

    def deserialize_field(self, Field field, model_instance, obj, list callbacks, dict key_to_instance):
        cdef str tag
        cdef object val
        cdef bint has_val
        cdef OneOfField oneoffield

        if isinstance(field, ReferenceField):
            tag = field.tag_name(self)
            val = obj.get(tag)
            if val is None:
                return
            setattr(model_instance, field.attr,
                    self._on_deserial(field, self._deserialize_reference(
                        field.type.model, val, callbacks, key_to_instance)))
        elif isinstance(field, OneOfField):
            oneoffield = <OneOfField>field
            has_val = False
            for f in oneoffield.fields:
                tag = f.tag_name(self)
                val = obj.get(tag)
                if val:
                    if not has_val:
                        has_val = True
                    else:
                        raise ValueError(
                            f'Only one of attributes({field.attrs}) can be specified')

                    setattr(model_instance, f.attr,
                            self._on_deserial(field, f.type.model.deserialize(self, obj[tag],
                                                                              callbacks, key_to_instance)))
        elif isinstance(field, ListField) and type(field.type.type) == Reference:
            tag = field.tag_name(self)
            if tag not in obj:
                return
            setattr(model_instance, field.attr,
                    self._on_deserial(
                        field, [self._deserialize_reference(field.type.type.model, it_obj, callbacks, key_to_instance)
                                if it_obj is not None else None
                                for it_obj in obj[tag]]))
        else:
            tag = field.tag_name(self)
            val = self._deserialize_value(obj.get(tag), callbacks, field.weak_ref)
            if val is None:
                return
            setattr(model_instance, field.attr, self._on_deserial(field, val))
            if isinstance(val, KeyPlaceholder):
                def cb(subs):
                    o = subs[val.key, val.id]
                    if field.weak_ref:
                        o = weakref.ref(o)
                    setattr(model_instance, field.attr, self._on_deserial(field, o))
                callbacks.append(cb)




# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import logging
import sys
from collections import defaultdict
from enum import Enum

from ... import promise
from ...config import options
from ...utils import classproperty, log_unhandled
from ...serialize import dataserializer
from .iorunner import IORunnerActor

logger = logging.getLogger(__name__)


class DataStorageDevice(Enum):
    PROC_MEMORY = 0
    CUDA = 1
    SHARED_MEMORY = 2
    DISK = 3
    VINEYARD = 4

    def __lt__(self, other):
        return self.value < other.value

    @classproperty
    def GLOBAL_DEVICES(cls):
        if options.vineyard.socket:
            return cls.VINEYARD, cls.DISK  # pragma: no cover
        else:
            return cls.SHARED_MEMORY, cls.DISK

    def build_location(self, proc_id):
        if self in self.GLOBAL_DEVICES:
            return 0, self
        else:
            return proc_id, self


class BytesStorageIO(object):
    storage_type = None

    def __init__(self, session_id, data_key, mode='w', handler=None, **kwargs):
        self._session_id = session_id
        self._data_key = data_key
        self._mode = mode
        self._buf = None
        self._handler = handler
        self._storage_ctx = handler.storage_ctx

        self._is_readable = 'r' in mode
        self._is_writable = 'w' in mode
        self._closed = False

    @property
    def is_readable(self):
        return self._is_readable and not self._closed

    @property
    def is_writable(self):
        return self._is_writable and not self._closed

    @property
    def closed(self):
        return self._closed

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close(finished=exc_val is None)

    def __del__(self):
        self.close()

    def get_io_pool(self, pool_name=None):
        return self._handler.get_io_pool(pool_name)

    def register(self, size, shape=None):
        location = self.storage_type.build_location(self._storage_ctx.proc_id)
        self._storage_ctx.manager_ref \
            .register_data(self._session_id, [self._data_key], location, [size], shapes=[shape])

    def close(self, finished=True):
        self._closed = True

    @property
    def nbytes(self):
        raise NotImplementedError

    def read(self, size=-1):
        raise NotImplementedError

    def write(self, d):
        raise NotImplementedError


class StorageHandler(object):
    storage_type = None

    def __init__(self, storage_ctx, proc_id=None):
        self._storage_ctx = storage_ctx
        self._actor_ctx = storage_ctx.actor_ctx
        self._proc_id = proc_id if proc_id is not None else storage_ctx.host_actor.proc_id
        if self.is_device_global():
            self._proc_id = 0

        from ..dispatcher import DispatchActor
        self._dispatch_ref = self.actor_ref(DispatchActor.default_uid())

    def get_io_pool(self, pool_name=None):
        actor_obj = self.host_actor
        pool_var = f'_pool_{pool_name}_attr_{self.storage_type.value}'
        if getattr(actor_obj, pool_var, None) is None:
            setattr(actor_obj, pool_var, self._actor_ctx.threadpool(1))
        return getattr(actor_obj, pool_var)

    @classmethod
    def is_device_global(cls):
        return cls.storage_type in DataStorageDevice.GLOBAL_DEVICES

    def is_other_process(self):
        return not self.is_device_global() and self.proc_id != self._storage_ctx.proc_id

    @staticmethod
    def pass_on_exc(func, exc):
        func()
        raise exc[1].with_traceback(exc[2]) from None

    @property
    def storage_ctx(self):
        return self._storage_ctx

    @property
    def host_actor(self):
        return self._storage_ctx.host_actor

    @property
    def proc_id(self):
        return self._proc_id

    @property
    def location(self):
        return self.storage_type.build_location(self._proc_id)

    def is_io_runner(self):
        return getattr(self.host_actor, '_io_runner', False)

    def actor_ref(self, *args, **kwargs):
        return self._actor_ctx.actor_ref(*args, **kwargs)

    def promise_ref(self, *args, **kwargs):
        return self._storage_ctx.host_actor.promise_ref(self.actor_ref(*args, **kwargs))

    def delete(self, session_id, data_keys, _tell=False):
        raise NotImplementedError

    def load_from(self, session_id, data_keys, src_handler, pin_token=None):
        """
        :param session_id: session id
        :param data_keys: keys of data to load
        :param src_handler: source data handler containing source data
        :type src_handler: StorageHandler
        :param pin_token: if not None, this token is used to pin loaded data.
                          pass this to unpin_data_keys() if unpin is needed
        """
        logger.debug('Try loading data %s from device %s into %s',
                     data_keys, src_handler.storage_type, self.storage_type)
        left_has_obj_io = getattr(self, '_has_object_io', False)
        right_has_obj_io = getattr(src_handler, '_has_object_io', False)
        right_has_bytes_io = getattr(src_handler, '_has_bytes_io', False)

        if left_has_obj_io and right_has_obj_io:
            return self.load_from_object_io(session_id, data_keys, src_handler, pin_token=pin_token)
        elif right_has_bytes_io:
            return self.load_from_bytes_io(session_id, data_keys, src_handler, pin_token=pin_token)
        else:
            return self.load_from_object_io(session_id, data_keys, src_handler, pin_token=pin_token)

    def load_from_bytes_io(self, session_id, data_keys, src_handler, pin_token=None):
        raise NotImplementedError

    def load_from_object_io(self, session_id, data_keys, src_handler, pin_token=None):
        raise NotImplementedError

    def register_data(self, session_id, data_keys, sizes, shapes=None):
        self._storage_ctx.manager_ref \
            .register_data(session_id, data_keys, self.location, sizes, shapes=shapes)

    def _dispatch_keys_to_uids(self, session_id, data_keys):
        uid_to_keys = defaultdict(list)
        runner_uids = self._dispatch_ref.get_hash_slots(
            'iorunner', [(session_id, k) for k in data_keys])
        for k, uid in zip(data_keys, runner_uids):
            uid_to_keys[uid].append(k)
        return uid_to_keys.items()

    def transfer_in_runner(self, session_id, data_keys, src_handler, fallback=None):
        if self.is_io_runner():
            return fallback() if fallback is not None else promise.finished()

        if self.is_device_global() and src_handler.is_device_global():
            return promise.all_(
                self.promise_ref(uid).load_from(
                    self.location, session_id, keys, src_handler.location, _promise=True)
                for uid, keys in self._dispatch_keys_to_uids(session_id, data_keys))
        elif self.is_device_global() or src_handler.is_device_global():
            if self.is_other_process() or src_handler.is_other_process():
                runner_proc_id = self.proc_id or src_handler.proc_id
                runner_ref = self.promise_ref(IORunnerActor.gen_uid(runner_proc_id))
                return runner_ref.load_from(
                    self.location, session_id, data_keys, src_handler.location, _promise=True)
            elif fallback is not None:
                if src_handler.storage_type != DataStorageDevice.DISK and \
                        self.storage_type != DataStorageDevice.DISK:
                    return fallback()

                uid_to_work_item_ids = dict()

                def _fallback_runner(uid, work_item_id):
                    uid_to_work_item_ids[uid] = work_item_id
                    return fallback()

                def _unlocker(uid, *exc, **kwargs):
                    self.promise_ref(uid).unlock(uid_to_work_item_ids[uid])
                    if not kwargs.get('accept', True):
                        raise exc[1].with_traceback(exc[2])

                return promise.all_(
                    self.promise_ref(uid).lock(session_id, keys, _promise=True)
                        .then(functools.partial(_fallback_runner, uid))
                        .then(functools.partial(_unlocker, uid), functools.partial(_unlocker, uid, accept=False))
                    for uid, keys in self._dispatch_keys_to_uids(session_id, data_keys)
                )
            else:
                return promise.finished()
        else:
            return fallback() if fallback is not None else promise.finished()

    def unregister_data(self, session_id, data_keys, _tell=False):
        self._storage_ctx.manager_ref \
            .unregister_data(session_id, data_keys, self.location, _tell=_tell)


class BytesStorageMixin(object):
    _has_bytes_io = True

    def create_bytes_reader(self, session_id, data_key, packed=False, packed_compression=None,
                            _promise=False):
        raise NotImplementedError

    def create_bytes_writer(self, session_id, data_key, total_bytes, packed=False,
                            packed_compression=None, auto_register=True, pin_token=None,
                            _promise=False):
        raise NotImplementedError

    def _copy_bytes_data(self, reader, writer, on_close=None):
        copy_block_size = options.worker.copy_block_size
        async_read_pool = reader.get_io_pool()
        async_write_pool = writer.get_io_pool()

        @log_unhandled
        def _copy(_reader, _writer):
            with_exc = False
            block = None
            write_future = None
            try:
                while True:
                    block = async_read_pool.submit(_reader.read, copy_block_size).result()
                    if write_future:
                        write_future.result()
                    if not block:
                        break
                    write_future = async_write_pool.submit(_writer.write, block)
            except:  # noqa: E722
                with_exc = True
                raise
            finally:
                if on_close:
                    on_close(_reader, _writer, not with_exc)
                _reader.close()
                _writer.close(finished=not with_exc)
                del _reader, _writer, block

        try:
            return self.host_actor.spawn_promised(_copy, reader, writer)
        finally:
            del reader, writer

    def _copy_object_data(self, serialized_obj, writer):
        def _copy(ser):
            try:
                with writer:
                    async_write_pool = writer.get_io_pool()
                    if hasattr(ser, 'write_to'):
                        async_write_pool.submit(ser.write_to, writer).result()
                    else:
                        async_write_pool.submit(writer.write, ser).result()
            finally:
                del ser

        try:
            return self.host_actor.spawn_promised(_copy, serialized_obj)
        finally:
            del serialized_obj


class ObjectStorageMixin(object):
    _has_object_io = True

    @staticmethod
    def _deserial(obj):
        if hasattr(obj, 'deserialize'):
            return obj.deserialize()
        else:
            return dataserializer.deserialize(obj)

    def _batch_load_objects(self, session_id, data_keys, key_loader,
                            batch_get=False, serialize=False, pin_token=None):
        is_success = True
        data_dict = dict()
        sizes = self._storage_ctx.manager_ref.get_data_sizes(session_id, data_keys)

        def _record_data(k, o):
            try:
                if is_success:
                    data_dict[k] = o
            finally:
                del o

        def _put_objects(*_):
            keys, objs = zip(*data_dict.items())
            data_dict.clear()
            key_list, obj_list = list(keys), list(objs)
            try:
                return self.put_objects(session_id, key_list, obj_list, sizes, serialize=serialize,
                                        pin_token=pin_token, _promise=True)
            finally:
                del objs
                obj_list[:] = []

        def _handle_err(*exc_info):
            nonlocal is_success
            is_success = False
            data_dict.clear()
            raise exc_info[1].with_traceback(exc_info[2]) from None

        def _batch_put_objects(objs):
            try:
                return self.put_objects(session_id, data_keys, objs, sizes,
                                        serialize=serialize, pin_token=pin_token, _promise=True)
            finally:
                del objs

        if batch_get:
            return key_loader(data_keys).then(_batch_put_objects)
        else:
            return promise.all_(
                key_loader(k).then(functools.partial(_record_data, k), _handle_err)
                for k in data_keys).then(_put_objects, _handle_err)

    def get_objects(self, session_id, data_keys, serialize=False, _promise=False):
        raise NotImplementedError

    def put_objects(self, session_id, data_keys, objs, sizes=None, shapes=None,
                    serialize=False, pin_token=False, _promise=False):
        raise NotImplementedError


class SpillableStorageMixin(object):
    _spillable = True

    def spill_size(self, size, multiplier=1):
        raise NotImplementedError

    def lift_data_keys(self, session_id, data_keys):
        raise NotImplementedError

    def pin_data_keys(self, session_id, data_keys, token):
        raise NotImplementedError

    def unpin_data_keys(self, session_id, data_keys, token, _tell=False):
        raise NotImplementedError


def wrap_promised(func):
    @functools.wraps(func)
    def _wrapped(*args, **kwargs):
        try:
            promised = kwargs.get('_promise')
            if not promised:
                return func(*args, **kwargs)
            else:
                try:
                    val = func(*args, **kwargs)
                    if isinstance(val, promise.Promise):
                        return val
                    else:
                        return promise.finished(val)
                except:  # noqa: E722
                    return promise.finished(*sys.exc_info(), _accept=False)
        finally:
            del args, kwargs

    return _wrapped


_storage_handler_cls = {}


def register_storage_handler_cls(storage_type, handler_cls):
    _storage_handler_cls[storage_type] = handler_cls


def get_storage_handler_cls(storage_type):
    try:
        return _storage_handler_cls[storage_type]
    except KeyError:  # pragma: no cover
        raise NotImplementedError(f'Storage type {storage_type} not supported')



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import itertools

import numpy as np

from ...config import options
from ...serialize import ValueType, StringField, TupleField
from ..utils import normalize_shape, decide_chunk_sizes
from ..core import TensorOrder
from ..operands import TensorOperand, TensorOperandMixin


class TensorDataSource(TensorOperand, TensorOperandMixin):
    """
    Tensor data source base class, provide universal tile logic,
    subclass can overwrite tile method.
    """

    __slots__ = ()

    def to_chunk_op(self, *args):
        chunk_shape = args[0]
        chunk_op = self.copy().reset_key()
        chunk_op.extra_params = {'size': chunk_shape}  # to make op key different
        return chunk_op

    @classmethod
    def tile(cls, op):
        tensor = op.outputs[0]

        chunk_size = tensor.extra_params.raw_chunk_size or options.chunk_size
        chunk_size = decide_chunk_sizes(tensor.shape, chunk_size, tensor.dtype.itemsize)
        chunk_size_idxes = (range(len(size)) for size in chunk_size)

        out_chunks = []
        for chunk_shape, chunk_idx in zip(itertools.product(*chunk_size),
                                          itertools.product(*chunk_size_idxes)):
            chunk_op = op.to_chunk_op(chunk_shape, chunk_idx, chunk_size)
            out_chunk = chunk_op.new_chunk(None, shape=chunk_shape, index=chunk_idx,
                                           order=tensor.order)
            out_chunks.append(out_chunk)

        new_op = op.copy()
        return new_op.new_tensors(op.inputs, tensor.shape, chunks=out_chunks, nsplits=chunk_size,
                                  order=tensor.order, **tensor.extra_params)


class TensorNoInput(TensorDataSource):
    """
    Tensor operand with no inputs.
    """

    def check_inputs(self, inputs):
        # no inputs
        if inputs and len(inputs) > 0:
            raise ValueError("Tensor data source has no inputs")

    def _new_chunks(self, inputs, kws=None, **kw):
        shape = kw.get('shape', None)
        self.extra_params['shape'] = shape  # set shape to make the operand key different
        return super()._new_chunks(inputs, kws=kws, **kw)

    def _new_tileables(self, inputs, kws=None, **kw):
        shape = kw.get('shape', None)
        self.extra_params['shape'] = shape  # set shape to make the operand key different
        return super()._new_tileables(inputs, kws=kws, **kw)

    def __call__(self, shape, chunk_size=None, order=None):
        shape = normalize_shape(shape)
        order = TensorOrder.C_ORDER if order is None else order
        return self.new_tensor(None, shape, raw_chunk_size=chunk_size, order=order)


class TensorHasInput(TensorDataSource):
    """
    Tensor operand with a single input.
    """

    @property
    def input(self):
        return self._input

    def check_inputs(self, inputs):
        # no inputs
        if len(inputs) != 1:
            raise ValueError("Tensor can only have 1 input")

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._input = self._inputs[0]

    @classmethod
    def tile(cls, op):
        output = op.outputs[0]

        out_chunks = []
        for c in op.input.chunks:
            out_chunk = op.copy().reset_key().new_chunk([c], shape=c.shape,
                                                        index=c.index, order=output.order)
            out_chunks.append(out_chunk)

        new_op = op.copy()
        return new_op.new_tensors(op.inputs, output.shape, order=output.order,
                                  chunks=out_chunks, nsplits=op.input.nsplits)

    def __call__(self, a, order=None):
        order = a.order if order is None else order
        return self.new_tensor([a], a.shape, order=order)


class TensorLike(TensorHasInput):
    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        if self.dtype is None:
            self._dtype = self.input.dtype
        if self.gpu is None:
            self._gpu = self.input.op.gpu

        # FIXME: remove when cupy supports other dtypes
        if self._gpu and self._dtype not in (np.float32, np.float64):
            raise NotImplementedError('Sparse tensor on GPU only supports float32 and float64')


class TensorFromHDF5Like(TensorNoInput):
    _filename = StringField('filename')
    _group = StringField('group')
    _dataset = StringField('dataset')
    _axis_offsets = TupleField('axis_offsets', ValueType.int64)

    def __init__(self, filename=None, group=None, dataset=None,
                 dtype=None, **kw):
        super().__init__(_filename=filename, _group=group,
                         _dataset=dataset, _dtype=dtype, **kw)

    @property
    def filename(self):
        return self._filename

    @property
    def group(self):
        return self._group

    @property
    def dataset(self):
        return self._dataset

    @property
    def axis_offsets(self):
        return self._axis_offsets

    @property
    def path(self):
        return self.get_path(self.group, self.dataset)

    def to_chunk_op(self, *args):
        _, chunk_index, nsplits = args
        chunk_op = super().to_chunk_op(*args)
        cum_offsets = [[0] + np.cumsum(ns).tolist() for ns in nsplits]
        axis_offsets = []
        for axis, idx in enumerate(chunk_index):
            axis_offsets.append(cum_offsets[axis][idx])
        chunk_op._axis_offsets = tuple(axis_offsets)
        return chunk_op

    @staticmethod
    def get_path(group, dataset):
        paths = []
        if group:
            paths.append(group)
        paths.append(dataset)
        return '/'.join(paths)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ..utils import infer_dtype
from .core import TensorUnaryOp
from .utils import arithmetic_operand


@arithmetic_operand(sparse_mode='unary')
class TensorFix(TensorUnaryOp):
    _op_type_ = OperandDef.FIX
    _func_name = 'fix'


@infer_dtype(np.fix)
def fix(x, out=None, **kwargs):
    """
    Round to nearest integer towards zero.

    Round a tensor of floats element-wise to nearest integer towards zero.
    The rounded values are returned as floats.

    Parameters
    ----------
    x   : array_like
        An tensor of floats to be rounded
    out : Tensor, optional
        Output tensor

    Returns
    -------
    out : Tensor of floats
        The array of rounded numbers

    See Also
    --------
    trunc, floor, ceil
    around : Round to given number of decimals

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.fix(3.14).execute()
    3.0
    >>> mt.fix(3).execute()
    3.0
    >>> mt.fix([2.1, 2.9, -2.1, -2.9]).execute()
    array([ 2.,  2., -2., -2.])

    """
    op = TensorFix(**kwargs)
    return op(x, out=out)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
try:
    import torch
    from torch.utils.data import Dataset
except ImportError:  # pragma: no cover
    torch = None
    Dataset = object

from ....context import get_context, RunningMode
from ....tensor.core import TENSOR_TYPE
from ....tensor.indexing.core import process_index
from ....dataframe.indexing.iloc import process_iloc_indexes
from ....utils import require_not_none


@require_not_none(torch)
class MarsDataset(Dataset):
    def __init__(self, *tileables):
        from ....session import Session

        self._context = get_context() or Session.default_or_local().context

        self.tileables = tileables
        self._datas = None
        self._offset = 0

    def prefetch(self, indices):
        self._datas = self._get_data(indices)
        self._offset = 0

    @staticmethod
    def _process_index(t, index):
        if isinstance(t, TENSOR_TYPE):
            return process_index(t.ndim, index)
        else:
            return process_iloc_indexes(t, index)

    def _get_data(self, item):
        if self._context.running_mode != RunningMode.distributed:
            return tuple(t[item].fetch() for t in self.tileables)
        else:
            return tuple(self._context.get_tileable_data(
                t.key, self._process_index(t, item)) for t in self.tileables)

    def __len__(self):
        return self.tileables[0].shape[0]

    def __getitem__(self, item):
        if self._datas is not None:
            ret = []
            for data in self._datas:
                if isinstance(data, np.ndarray):
                    ret.append(data[self._offset])
                else:
                    ret.append(data.iloc[self._offset])
            self._offset += 1
            return ret
        else:
            return self._get_data(item)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ..utils import infer_dtype
from .core import TensorBinOp
from .utils import arithmetic_operand


@arithmetic_operand
class TensorFMin(TensorBinOp):
    _op_type_ = OperandDef.FMIN
    _func_name = 'fmin'

    @classmethod
    def _is_sparse(cls, x1, x2):
        if hasattr(x1, 'issparse') and x1.issparse() and np.isscalar(x2) and x2 >= 0:
            return True
        if hasattr(x2, 'issparse') and x2.issparse() and np.isscalar(x1) and x1 >= 0:
            return True
        return False


@infer_dtype(np.fmin)
def fmin(x1, x2, out=None, where=None, **kwargs):
    """
    Element-wise minimum of array elements.

    Compare two tensors and returns a new tensor containing the element-wise
    minima. If one of the elements being compared is a NaN, then the
    non-nan element is returned. If both elements are NaNs then the first
    is returned.  The latter distinction is important for complex NaNs,
    which are defined as at least one of the real or imaginary parts being
    a NaN. The net effect is that NaNs are ignored when possible.

    Parameters
    ----------
    x1, x2 : array_like
        The tensors holding the elements to be compared. They must have
        the same shape.
    out : Tensor, None, or tuple of Tensor and None, optional
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or `None`,
        a freshly-allocated tensor is returned. A tuple (possible only as a
        keyword argument) must have length equal to the number of outputs.
    where : array_like, optional
        Values of True indicate to calculate the ufunc at that position, values
        of False indicate to leave the value in the output alone.
    **kwargs

    Returns
    -------
    y : Tensor or scalar
        The minimum of `x1` and `x2`, element-wise.  Returns scalar if
        both  `x1` and `x2` are scalars.

    See Also
    --------
    fmax :
        Element-wise maximum of two tensors, ignores NaNs.
    minimum :
        Element-wise minimum of two tensors, propagates NaNs.
    amin :
        The minimum value of a tensor along a given axis, propagates NaNs.
    nanmin :
        The minimum value of a tensor along a given axis, ignores NaNs.

    maximum, amax, nanmax

    Notes
    -----

    The fmin is equivalent to ``mt.where(x1 <= x2, x1, x2)`` when neither
    x1 nor x2 are NaNs, but it is faster and does proper broadcasting.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.fmin([2, 3, 4], [1, 5, 2]).execute()
    array([1, 3, 2])

    >>> mt.fmin(mt.eye(2), [0.5, 2]).execute()
    array([[ 0.5,  0. ],
           [ 0. ,  1. ]])

    >>> mt.fmin([mt.nan, 0, mt.nan],[0, mt.nan, mt.nan]).execute()
    array([  0.,   0.,  NaN])
    """
    op = TensorFMin(**kwargs)
    return op(x1, x2, out=out, where=where)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from sklearn.utils.extmath import row_norms as sklearn_row_norms

from ... import opcodes
from ...operands import OutputType, OperandStage
from ...serialize import KeyField, BoolField, Int32Field
from ...tensor.array_utils import as_same_device, device, sparse
from ...tensor.core import TensorOrder
from ...tiles import TilesError
from ...utils import check_chunks_unknown_shape
from ..operands import LearnOperand, LearnOperandMixin
from ._k_means_common import _execute_merge_update, _relocate_empty_clusters
from ._k_means_fast import update_center
from ._k_means_lloyd import update_chunk_dense, update_chunk_sparse


class KMeansLloydUpdate(LearnOperand, LearnOperandMixin):
    _op_type_ = opcodes.KMEANS_LLOYD_UPDATE

    _x = KeyField('x')
    _sample_weight = KeyField('sample_weight')
    _x_squared_norms = KeyField('x_squared_norms')
    _centers_old = KeyField('centers_old')
    _labels = KeyField('labels')
    _update_centers = BoolField('update_centers')
    _n_clusters = Int32Field('n_clusters')

    def __init__(self, x=None, sample_weight=None, x_squared_norms=None,
                 centers_old=None, labels=None, update_centers=None,
                 n_clusters=None, output_types=None, stage=None, **kw):
        super().__init__(_x=x, _sample_weight=sample_weight,
                         _x_squared_norms=x_squared_norms,
                         _centers_old=centers_old, _labels=labels,
                         _update_centers=update_centers, _n_clusters=n_clusters,
                         _output_types=output_types, _stage=stage, **kw)
        if self._output_types is None:
            self._output_types = [OutputType.tensor] * self.output_limit

    @property
    def x(self):
        return self._x

    @property
    def sample_weight(self):
        return self._sample_weight

    @property
    def x_squared_norms(self):
        return self._x_squared_norms

    @property
    def centers_old(self):
        return self._centers_old

    @property
    def labels(self):
        return self._labels

    @property
    def update_centers(self):
        return self._update_centers

    @property
    def n_clusters(self):
        return self._n_clusters

    @property
    def output_limit(self):
        return 3 if self._stage != OperandStage.reduce else 2

    @property
    def _input_fields(self):
        return '_x', '_sample_weight', '_x_squared_norms', \
               '_centers_old', '_labels'

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        inputs_iter = iter(self._inputs)
        for field in self._input_fields:
            if getattr(self, field, None) is not None:
                setattr(self, field, next(inputs_iter))

    def __call__(self):
        kws = [
            # labels
            self._labels.params,
            # centers_new
            {
                'shape': (self._n_clusters, self._x.shape[1]),
                'dtype': self._centers_old.dtype,
                'order': TensorOrder.C_ORDER
            },
            # weight_in_clusters
            {
                'shape': (self._n_clusters,),
                'dtype': self._centers_old.dtype,
                'order': TensorOrder.C_ORDER
            }
        ]
        return self.new_tileables(
            [getattr(self, field) for field in self._input_fields], kws=kws)

    @classmethod
    def tile(cls, op: "KMeansLloydUpdate"):
        check_chunks_unknown_shape(op.inputs, TilesError)
        x = op.x
        if x.chunk_shape[1] != 1:  # pragma: no cover
            x = x.rechunk({1: x.shape[1]})._inplace_tile()
        sample_weight = op.sample_weight.rechunk({0: x.nsplits[0]})._inplace_tile()
        x_squared_norms = op.x_squared_norms.rechunk({0: x.nsplits[0]})._inplace_tile()
        labels = op.labels.rechunk({0: x.nsplits[0]})._inplace_tile()
        assert len(op.centers_old.chunks) == 1

        labels_chunks, centers_new_chunks, weight_in_clusters_chunks = [], [], []
        for i in range(x.chunk_shape[0]):
            x_chunk = x.cix[i, 0]
            sample_weight_chunk = sample_weight.cix[i, ]
            x_squared_norms_chunk = x_squared_norms.cix[i, ]
            labels_chunk = labels.cix[i, ]
            chunk_op = op.copy().reset_key()
            chunk_op._stage = OperandStage.map
            chunk_kws = [
                labels_chunk.params,
                {'index': (0, 0),
                 'shape': (op.n_clusters, x_chunk.shape[1]),
                 'dtype': op.centers_old.dtype,
                 'order': TensorOrder.C_ORDER},
                {'index': (0,),
                 'shape': (op.n_clusters,),
                 'dtype': op.centers_old.dtype,
                 'order': TensorOrder.C_ORDER}
            ]
            labels_chunk, centers_new_chunk, weight_in_clusters_chunk = chunk_op.new_chunks(
                [x_chunk, sample_weight_chunk, x_squared_norms_chunk,
                 op.centers_old.chunks[0], labels_chunk], kws=chunk_kws)
            labels_chunks.append(labels_chunk)
            centers_new_chunks.append(centers_new_chunk)
            weight_in_clusters_chunks.append(weight_in_clusters_chunk)

        if op.update_centers:
            # merge centers_new and weight_in_clusters
            merge_op = KMeansLloydUpdate(stage=OperandStage.reduce,
                                         n_clusters=op.n_clusters)
            merge_chunk_kw = [
                centers_new_chunks[0].params,
                weight_in_clusters_chunks[0].params
            ]
            centers_new_chunk, weight_in_cluster_chunk = merge_op.new_chunks(
                centers_new_chunks + weight_in_clusters_chunks, kws=merge_chunk_kw)
        else:
            # the data is meaningless, just pick one
            centers_new_chunk = centers_new_chunks[0]
            weight_in_cluster_chunk = weight_in_clusters_chunks[0]

        out_params = [out.params for out in op.outputs]
        # labels
        out_params[0]['nsplits'] = labels.nsplits
        out_params[0]['chunks'] = labels_chunks
        # centers_new
        out_params[1]['nsplits'] = tuple((s,) for s in op.outputs[1].shape)
        out_params[1]['chunks'] = [centers_new_chunk]
        # weight_in_clusters
        out_params[2]['nsplits'] = tuple((s,) for s in op.outputs[2].shape)
        out_params[2]['chunks'] = [weight_in_cluster_chunk]
        new_op = op.copy()
        return new_op.new_tileables(op.inputs, kws=out_params)

    @classmethod
    def _execute_reduce(cls, ctx, op):
        return _execute_merge_update(ctx, op)

    @classmethod
    def execute(cls, ctx, op: "KMeansLloydUpdate"):
        if op.stage == OperandStage.reduce:
            return cls._execute_reduce(ctx, op)
        else:
            (x, sample_weight, x_squared_norms, centers_old, labels), device_id, xp = as_same_device(
                [ctx[inp.key] for inp in op.inputs], device=op.device,
                ret_extra=True, copy_if_not_writeable=True)

            with device(device_id):
                if not op.update_centers:
                    centers_new = centers_old.copy()
                else:
                    centers_new = np.zeros_like(centers_old)
                weight_in_clusters = np.zeros(op.n_clusters, dtype=x.dtype)
                centers_squared_norms = sklearn_row_norms(centers_old, squared=True)

                if xp is np:
                    method = update_chunk_dense
                elif xp is sparse:
                    method = update_chunk_sparse
                else:  # pragma: no cover
                    raise NotImplementedError('Does not support run on GPU')
                out_labels = labels.copy()
                method(x, sample_weight, x_squared_norms, centers_old,
                       centers_squared_norms, out_labels, centers_new,
                       weight_in_clusters, op.update_centers)

                # labels
                ctx[op.outputs[0].key] = out_labels
                # centers_new
                ctx[op.outputs[1].key] = centers_new
                # weight_in_cluster
                ctx[op.outputs[2].key] = weight_in_clusters


class KMeansLloydPostprocess(LearnOperand, LearnOperandMixin):
    _op_type_ = opcodes.KMEANS_LLOYD_POSTPROCESS

    _centers_old = KeyField('centers_old')
    _centers_new = KeyField('centers_new')
    _center_shift = KeyField('center_shift')
    _weight_in_clusters = KeyField('weight_in_clusters')

    def __init__(self, centers_old=None, centers_new=None,
                 center_shift=None, weight_in_clusters=None,
                 output_types=None, stage=None, **kw):
        super().__init__(_centers_old=centers_old, _centers_new=centers_new,
                         _center_shift=center_shift,
                         _weight_in_clusters=weight_in_clusters,
                         _output_types=output_types, _stage=stage, **kw)
        if self._output_types is None:
            self._output_types = [OutputType.tensor] * self.output_limit

    @property
    def centers_old(self):
        return self._centers_old

    @property
    def centers_new(self):
        return self._centers_new

    @property
    def center_shift(self):
        return self._center_shift

    @property
    def weight_in_clusters(self):
        return self._weight_in_clusters

    @property
    def output_limit(self):
        return 2

    @property
    def _input_fields(self):
        return '_centers_old', '_centers_new', '_center_shift', \
               '_weight_in_clusters'

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        inputs_iter = iter(self._inputs)
        for field in self._input_fields:
            ob = getattr(self, field)
            if ob is not None:
                setattr(self, field, next(inputs_iter))

    def __call__(self):
        kws = [
            # centers_new
            self._centers_new.params,
            # center_shift
            self._center_shift.params
        ]
        return self.new_tileables(
            [getattr(self, f) for f in self._input_fields], kws=kws)

    @classmethod
    def tile(cls, op: "KMeansLloydPostprocess"):
        assert len(op.centers_old.chunks) == 1
        assert len(op.centers_new.chunks) == 1
        assert len(op.center_shift.chunks) == 1
        assert len(op.weight_in_clusters.chunks) == 1

        centers_old_chunk = op.centers_old.chunks[0]
        centers_new_chunk = op.centers_new.chunks[0]
        center_shift_chunk = op.center_shift.chunks[0]
        weight_in_clusters_chunk = op.weight_in_clusters.chunks[0]
        centers_new_chunk, center_shift_chunk = KMeansLloydPostprocess(
            centers_old=centers_old_chunk,
            centers_new=centers_new_chunk,
            center_shift=center_shift_chunk,
            weight_in_clusters=weight_in_clusters_chunk
        ).new_chunks([centers_old_chunk, centers_new_chunk,
                      center_shift_chunk, weight_in_clusters_chunk],
                     kws=[centers_new_chunk.params, center_shift_chunk.params])

        centers_new_kw = op.centers_new.params
        centers_new_kw['chunks'] = [centers_new_chunk]
        centers_new_kw['nsplits'] = op.centers_new.nsplits
        center_shift_kw = op.center_shift.params
        center_shift_kw['chunks'] = [center_shift_chunk]
        center_shift_kw['nsplits'] = op.center_shift.nsplits
        new_op = op.copy()
        return new_op.new_tileables(
            op.inputs, kws=[centers_new_kw, center_shift_kw])

    @classmethod
    def execute(cls, ctx, op: "KMeansLloydPostprocess"):
        (centers_old, centers_new, center_shift, weight_in_clusters), device_id, xp = \
            as_same_device([ctx[inp.key] for inp in op.inputs], op.device,
                           ret_extra=True, copy_if_not_writeable=True)

        with device(device_id):
            out_center_shift = center_shift.copy()
            out_centers_new = centers_new.copy()
            update_center(centers_old, out_centers_new,
                          out_center_shift, weight_in_clusters)

            ctx[op.outputs[0].key] = out_centers_new
            ctx[op.outputs[1].key] = out_center_shift


def lloyd_iter(X, sample_weight, x_squared_norms, centers_old, labels,
               center_shift, update_centers=True, session=None, run_kwargs=None):
    update_op = KMeansLloydUpdate(x=X, sample_weight=sample_weight,
                                  x_squared_norms=x_squared_norms,
                                  centers_old=centers_old, labels=labels,
                                  update_centers=update_centers,
                                  n_clusters=centers_old.shape[0])
    to_run = []
    ret = update_op()
    to_run.extend(ret)
    labels, centers_new, weight_in_clusters = ret

    if update_centers:
        centers_new, weight_in_clusters = \
            _relocate_empty_clusters(X, sample_weight, centers_old, centers_new,
                                     weight_in_clusters, labels, to_run=to_run,
                                     session=session, run_kwargs=run_kwargs)
        postprocess = KMeansLloydPostprocess(
            centers_old=centers_old, centers_new=centers_new,
            center_shift=center_shift, weight_in_clusters=weight_in_clusters)
        centers_new, center_shift = postprocess()

    return centers_new, weight_in_clusters, labels, center_shift



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ...serialize import KeyField
from ..operands import DataFrameOperand, DataFrameOperandMixin
from ..core import DATAFRAME_TYPE, SERIES_TYPE


class DataFrameDeviceConversionBase(DataFrameOperand, DataFrameOperandMixin):
    _input = KeyField('input')

    @property
    def input(self):
        return self._input

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._input = inputs[0]

    def __call__(self, obj):
        if isinstance(obj, DATAFRAME_TYPE):
            return self.new_dataframe([obj], shape=obj.shape, dtypes=obj.dtypes,
                                      index_value=obj.index_value,
                                      columns_value=obj.columns_value)
        else:
            assert isinstance(obj, SERIES_TYPE)
            return self.new_series([obj], shape=obj.shape, dtype=obj.dtype,
                                   index_value=obj.index_value, name=obj.name)

    @classmethod
    def tile(cls, op):
        out_chunks = []
        for c in op.input.chunks:
            chunk_op = op.copy().reset_key()
            out_chunk = chunk_op.new_chunk([c], **c.params)
            out_chunks.append(out_chunk)

        new_op = op.copy().reset_key()
        out = op.outputs[0]
        return new_op.new_tileables(op.inputs, chunks=out_chunks,
                                    nsplits=op.inputs[0].nsplits,
                                    **out.params)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import binascii
import datetime
import pickle
import uuid

import numpy as np
import pandas as pd
import cloudpickle

from ... import opcodes as OperandDef
from ...config import options
from ...serialize import StringField, AnyField, BoolField, ListField, \
    Int64Field, Float64Field, BytesField
from ...tensor.utils import normalize_chunk_sizes
from ..arrays import ArrowStringDtype
from ..core import IndexValue
from ..operands import DataFrameOperand, DataFrameOperandMixin, OutputType
from ..utils import parse_index, create_sa_connection, \
    standardize_range_index, to_arrow_dtypes


class DataFrameReadSQL(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = OperandDef.READ_SQL

    _table_or_sql = StringField('table_or_sql')
    _selectable = BytesField('selectable', on_serialize=pickle.dumps,
                             on_deserialize=pickle.loads)
    _con = StringField('con')
    _schema = StringField('schema')
    _index_col = AnyField('index_col')
    _coerce_float = BoolField('coerce_float')
    _parse_dates = AnyField('parse_dates')
    _columns = ListField('columns')
    _engine_kwargs = BytesField('engine_kwargs', on_serialize=cloudpickle.dumps,
                                on_deserialize=cloudpickle.loads)
    _row_memory_usage = Float64Field('row_memory_usage')
    _method = StringField('method')
    _incremental_index = BoolField('incremental_index')
    _use_arrow_dtype = BoolField('use_arrow_dtype')
    # for chunks
    _offset = Int64Field('offset')
    _partition_col = StringField('partition_col')
    _num_partitions = Int64Field('num_partitions')
    _low_limit = AnyField('low_limit')
    _high_limit = AnyField('high_limit')
    _left_end = BoolField('left_end')
    _right_end = BoolField('right_end')
    _nrows = Int64Field('nrows')

    def __init__(self, table_or_sql=None, selectable=None, con=None, schema=None,
                 index_col=None, coerce_float=None, parse_dates=None, columns=None,
                 engine_kwargs=None, row_memory_usage=None, method=None,
                 incremental_index=None, use_arrow_dtype=None, offset=None, partition_col=None,
                 num_partitions=None, low_limit=None, high_limit=None, left_end=None,
                 right_end=None, nrows=None, output_types=None, gpu=None, **kw):
        super().__init__(_table_or_sql=table_or_sql, _selectable=selectable, _con=con,
                         _schema=schema, _index_col=index_col, _coerce_float=coerce_float,
                         _parse_dates=parse_dates, _columns=columns,
                         _engine_kwargs=engine_kwargs, _row_memory_usage=row_memory_usage,
                         _method=method, _incremental_index=incremental_index,
                         _use_arrow_dtype=use_arrow_dtype, _offset=offset,
                         _partition_col=partition_col, _num_partitions=num_partitions,
                         _low_limit=low_limit, _left_end=left_end, _right_end=right_end,
                         _high_limit=high_limit, _nrows=nrows, _output_types=output_types,
                         _gpu=gpu, **kw)
        if not self.output_types:
            self._output_types = [OutputType.dataframe]

    @property
    def table_or_sql(self):
        return self._table_or_sql

    @property
    def selectable(self):
        return self._selectable

    @property
    def con(self):
        return self._con

    @property
    def schema(self):
        return self._schema

    @property
    def index_col(self):
        return self._index_col

    @property
    def coerce_float(self):
        return self._coerce_float

    @property
    def parse_dates(self):
        return self._parse_dates

    @property
    def columns(self):
        return self._columns

    @property
    def engine_kwargs(self):
        return self._engine_kwargs

    @property
    def row_memory_usage(self):
        return self._row_memory_usage

    @property
    def method(self):
        return self._method

    @property
    def incremental_index(self):
        return self._incremental_index

    @property
    def use_arrow_dtype(self):
        return self._use_arrow_dtype

    @property
    def offset(self):
        return self._offset

    @property
    def partition_col(self):
        return self._partition_col

    @property
    def num_partitions(self):
        return self._num_partitions

    @property
    def low_limit(self):
        return self._low_limit

    @property
    def high_limit(self):
        return self._high_limit

    @property
    def left_end(self):
        return self._left_end

    @property
    def right_end(self):
        return self._right_end

    @property
    def nrows(self):
        return self._nrows

    def _get_selectable(self, engine_or_conn, columns=None):
        import sqlalchemy as sa
        from sqlalchemy import sql
        from sqlalchemy.exc import SQLAlchemyError

        # process table_name
        if self._selectable is not None:
            selectable = self._selectable
        else:
            if isinstance(self._table_or_sql, sa.Table):
                selectable = self._table_or_sql
                self._table_or_sql = selectable.name
            else:
                m = sa.MetaData()
                try:
                    selectable = sa.Table(self._table_or_sql, m, autoload=True,
                                          autoload_with=engine_or_conn, schema=self._schema)
                except SQLAlchemyError:
                    temp_name_1 = 't1_' + binascii.b2a_hex(uuid.uuid4().bytes).decode()
                    temp_name_2 = 't2_' + binascii.b2a_hex(uuid.uuid4().bytes).decode()
                    if columns:
                        selectable = sql.text(self._table_or_sql).columns(*[sql.column(c) for c in columns]) \
                            .alias(temp_name_2)
                    else:
                        selectable = sql.select(
                            '*', from_obj=sql.text(f'({self._table_or_sql}) AS {temp_name_1}')) \
                            .alias(temp_name_2)
                    self._selectable = selectable
        return selectable

    def _collect_info(self, engine_or_conn, selectable, columns, test_rows):
        from sqlalchemy import sql

        # fetch test DataFrame
        if columns:
            query = sql.select([sql.column(c) for c in columns], from_obj=selectable).limit(test_rows)
        else:
            query = sql.select(selectable.columns, from_obj=selectable).limit(test_rows)
        test_df = pd.read_sql(query, engine_or_conn, index_col=self._index_col,
                              coerce_float=self._coerce_float,
                              parse_dates=self._parse_dates)
        if len(test_df) == 0:
            self._row_memory_usage = None
        else:
            self._row_memory_usage = \
                test_df.memory_usage(deep=True, index=True).sum() / len(test_df)

        if self._method == 'offset':
            # fetch size
            size = list(engine_or_conn.execute(
                sql.select([sql.func.count()]).select_from(selectable)))[0][0]
            shape = (size, test_df.shape[1])
        else:
            shape = (np.nan, test_df.shape[1])

        return test_df, shape

    def __call__(self, test_rows, chunk_size):
        import sqlalchemy as sa
        from sqlalchemy.sql import elements

        with create_sa_connection(self._con, **(self._engine_kwargs or dict())) as con:
            self._con = str(con.engine.url)
            selectable = self._get_selectable(con)

            # process index_col
            index_col = self._index_col
            if index_col is not None:
                if not isinstance(index_col, (list, tuple)):
                    index_col = (index_col,)
                new_index_col = []
                for col in index_col:
                    if isinstance(col, (sa.Column, elements.Label)):
                        new_index_col.append(col.name)
                    elif isinstance(col, str):
                        new_index_col.append(col)
                    elif col is not None:
                        raise TypeError(f'unknown index_col type: {type(col)}')
                self._index_col = new_index_col

            # process columns
            columns = self._columns or []
            new_columns = []
            for col in columns:
                if isinstance(col, str):
                    new_columns.append(col)
                else:
                    new_columns.append(col.name)
            self._columns = new_columns

            if self._columns:
                collect_cols = self._columns + (self._index_col or [])
            else:
                collect_cols = []
            test_df, shape = self._collect_info(con, selectable, collect_cols, test_rows)

            # reconstruct selectable using known column names
            if not collect_cols:
                self._columns = list(test_df.columns)
                if self._selectable is not None:
                    self._selectable = None
                    self._get_selectable(con, columns=self._columns + (self._index_col or []))

            if self.method == 'partition':
                if not self.index_col or self.partition_col not in self.index_col:
                    part_frame = test_df
                else:
                    part_frame = test_df.index.to_frame()

                if not issubclass(part_frame[self.partition_col].dtype.type, (np.number, np.datetime64)):
                    raise TypeError('Type of partition column should be numeric or datetime, '
                                    f'now it is {test_df[self.partition_col].dtype}')

            if isinstance(test_df.index, pd.RangeIndex):
                index_value = parse_index(pd.RangeIndex(shape[0] if not np.isnan(shape[0]) else -1),
                                          str(selectable), self._con)
            else:
                index_value = parse_index(test_df.index)

            columns_value = parse_index(test_df.columns, store_data=True)

            dtypes = test_df.dtypes
            use_arrow_dtype = self._use_arrow_dtype
            if use_arrow_dtype is None:
                use_arrow_dtype = options.dataframe.use_arrow_dtype
            if use_arrow_dtype:
                dtypes = to_arrow_dtypes(dtypes, test_df=test_df)

            return self.new_dataframe(None, shape=shape, dtypes=dtypes,
                                      index_value=index_value,
                                      columns_value=columns_value,
                                      raw_chunk_size=chunk_size)

    @classmethod
    def _tile_offset(cls, op: 'DataFrameReadSQL'):
        df = op.outputs[0]

        if op.row_memory_usage is not None:
            # Data selected
            chunk_size = df.extra_params.raw_chunk_size or options.chunk_size
            if chunk_size is None:
                chunk_size = (int(options.chunk_store_limit / op.row_memory_usage), df.shape[1])
            row_chunk_sizes = normalize_chunk_sizes(df.shape, chunk_size)[0]
        else:
            # No data selected
            row_chunk_sizes = (0,)
        offsets = np.cumsum((0,) + row_chunk_sizes).tolist()

        out_chunks = []
        for i, row_size in enumerate(row_chunk_sizes):
            chunk_op = op.copy().reset_key()
            chunk_op._row_memory_usage = None  # no need for chunk
            offset = chunk_op._offset = offsets[i]
            if df.index_value.has_value():
                # range index
                index_value = parse_index(
                    df.index_value.to_pandas()[offset: offsets[i + 1]])
            else:
                index_value = parse_index(df.index_value.to_pandas(),
                                          op.table_or_sql or str(op.selectable), op.con, i, row_size)
            out_chunk = chunk_op.new_chunk(None, shape=(row_size, df.shape[1]),
                                           columns_value=df.columns_value,
                                           index_value=index_value, dtypes=df.dtypes,
                                           index=(i, 0))
            out_chunks.append(out_chunk)

        nsplits = (row_chunk_sizes, (df.shape[1],))
        new_op = op.copy()
        return new_op.new_dataframes(None, chunks=out_chunks, nsplits=nsplits,
                                     **df.params)

    def _parse_datetime(self, val):
        if isinstance(self.parse_dates, list):
            return pd.to_datetime(val)
        args = self.parse_dates[self.partition_col]
        args = {'format': args} if isinstance(args, str) else args
        return pd.to_datetime(val, **args)

    @classmethod
    def _tile_partition(cls, op: 'DataFrameReadSQL'):
        df = op.outputs[0]

        selectable = op._get_selectable(None)

        if op._low_limit is None or op._high_limit is None:
            import sqlalchemy as sa
            from sqlalchemy import sql

            engine = sa.create_engine(op.con, **(op.engine_kwargs or dict()))
            try:
                part_col = selectable.columns[op.partition_col]
                range_results = engine.execute(sql.select([sql.func.min(part_col), sql.func.max(part_col)]))

                op._low_limit, op._high_limit = next(range_results)
                if op.parse_dates and op.partition_col in op.parse_dates:
                    op._low_limit = op._parse_datetime(op._low_limit)
                    op._high_limit = op._parse_datetime(op._high_limit)
            finally:
                engine.dispose()

        if isinstance(op._low_limit, (datetime.datetime, np.datetime64, pd.Timestamp)):
            seps = pd.date_range(op._low_limit, op._high_limit, op.num_partitions + 1)
        else:
            seps = np.linspace(op._low_limit, op._high_limit, op.num_partitions + 1, endpoint=True)

        out_chunks = []
        for i, (start, end) in enumerate(zip(seps, seps[1:])):
            chunk_op = op.copy().reset_key()
            chunk_op._row_memory_usage = None  # no need for chunk
            chunk_op._num_partitions = None
            chunk_op._low_limit = start
            chunk_op._high_limit = end
            chunk_op._left_end = i == 0
            chunk_op._right_end = i == op.num_partitions - 1

            if df.index_value.has_value():
                # range index
                index_value = parse_index(-1, chunk_op.key, chunk_op.index_value.key)
            else:
                index_value = parse_index(df.index_value.to_pandas(),
                                          str(selectable), op.con, i)
            out_chunk = chunk_op.new_chunk(None, shape=(np.nan, df.shape[1]),
                                           columns_value=df.columns_value,
                                           index_value=index_value, dtypes=df.dtypes,
                                           index=(i, 0))
            out_chunks.append(out_chunk)

        if op.incremental_index and len(out_chunks) > 1 and \
                isinstance(df.index_value._index_value, IndexValue.RangeIndex):
            out_chunks = standardize_range_index(out_chunks)

        nsplits = ((np.nan,) * len(out_chunks), (df.shape[1],))
        new_op = op.copy()
        return new_op.new_dataframes(None, chunks=out_chunks, nsplits=nsplits,
                                     **df.params)

    @classmethod
    def tile(cls, op: 'DataFrameReadSQL'):
        if op.method == 'offset':
            return cls._tile_offset(op)
        else:
            return cls._tile_partition(op)

    @classmethod
    def execute(cls, ctx, op: 'DataFrameReadSQL'):
        import sqlalchemy as sa

        def _adapt_datetime(dt):
            if isinstance(dt, np.datetime64):
                return dt.astype('<M8[ms]').astype(datetime.datetime)
            elif isinstance(dt, pd.Timestamp):
                return dt.to_pydatetime()
            return dt

        out = op.outputs[0]

        engine = sa.create_engine(op.con, **(op.engine_kwargs or dict()))
        try:
            selectable = op._get_selectable(engine)

            columns = [selectable.columns[col] for col in op.columns]
            column_names = set(op.columns)
            if op.index_col:
                for icol in op.index_col:
                    if icol not in column_names:
                        columns.append(selectable.columns[icol])

            # convert to python timestamp in case np / pd time types not handled
            op._low_limit = _adapt_datetime(op._low_limit)
            op._high_limit = _adapt_datetime(op._high_limit)

            query = sa.sql.select(columns)
            if op.method == 'partition':
                part_col = selectable.columns[op.partition_col]
                if op.left_end:
                    query = query.where(part_col < op.high_limit)
                elif op.right_end:
                    query = query.where(part_col >= op.low_limit)
                else:
                    query = query.where((part_col >= op.low_limit) & (part_col < op.high_limit))

            if hasattr(selectable, 'primary_key') and len(selectable.primary_key) > 0:
                # if table has primary key, sort as the order
                query = query.order_by(*list(selectable.primary_key))
            elif op.index_col:
                # if no primary key, sort as the index_col
                query = query.order_by(
                    *[selectable.columns[col] for col in op.index_col])
            else:
                # at last, we sort by all the columns
                query = query.order_by(*columns)

            if op.method == 'offset':
                query = query.limit(out.shape[0])
                if op.offset > 0:
                    query = query.offset(op.offset)

            if op.nrows is not None:
                query = query.limit(op.nrows)

            df = pd.read_sql(query, engine, index_col=op.index_col,
                             coerce_float=op.coerce_float,
                             parse_dates=op.parse_dates)
            if op.method == 'offset' and op.index_col is None and op.offset > 0:
                index = pd.RangeIndex(op.offset, op.offset + out.shape[0])
                if op.nrows is not None:
                    index = index[:op.nrows]
                df.index = index

            use_arrow_dtype = op.use_arrow_dtype
            if use_arrow_dtype is None:
                use_arrow_dtype = options.dataframe.use_arrow_dtype
            if use_arrow_dtype:
                dtypes = to_arrow_dtypes(df.dtypes, test_df=df)
                for i in range(len(dtypes)):
                    dtype = dtypes.iloc[i]
                    if isinstance(dtype, ArrowStringDtype):
                        df.iloc[:, i] = df.iloc[:, i].astype(dtype)

            ctx[out.key] = df
        finally:
            engine.dispose()


def _read_sql(table_or_sql, con, schema=None, index_col=None, coerce_float=True,
              params=None, parse_dates=None, columns=None, chunksize=None,
              incremental_index=False, use_arrow_dtype=None,
              test_rows=None, chunk_size=None,
              engine_kwargs=None, partition_col=None, num_partitions=None,
              low_limit=None, high_limit=None):
    if chunksize is not None:
        raise NotImplementedError('read_sql_query with chunksize not supported')
    method = 'offset' if partition_col is None else 'partition'

    op = DataFrameReadSQL(table_or_sql=table_or_sql, selectable=None, con=con, schema=schema,
                          index_col=index_col, coerce_float=coerce_float,
                          params=params, parse_dates=parse_dates, columns=columns,
                          engine_kwargs=engine_kwargs, incremental_index=incremental_index,
                          use_arrow_dtype=use_arrow_dtype, method=method, partition_col=partition_col,
                          num_partitions=num_partitions, low_limit=low_limit,
                          high_limit=high_limit)
    return op(test_rows, chunk_size)


def read_sql(sql, con, index_col=None, coerce_float=True, params=None, parse_dates=None,
             columns=None, chunksize=None, test_rows=5, chunk_size=None, engine_kwargs=None,
             incremental_index=None, partition_col=None, num_partitions=None, low_limit=None,
             high_limit=None):
    """
    Read SQL query or database table into a DataFrame.

    This function is a convenience wrapper around ``read_sql_table`` and
    ``read_sql_query`` (for backward compatibility). It will delegate
    to the specific function depending on the provided input. A SQL query
    will be routed to ``read_sql_query``, while a database table name will
    be routed to ``read_sql_table``. Note that the delegated function might
    have more specific notes about their functionality not listed here.

    Parameters
    ----------
    sql : str or SQLAlchemy Selectable (select or text object)
        SQL query to be executed or a table name.
    con : SQLAlchemy connectable (engine/connection) or database str URI
        or DBAPI2 connection (fallback mode)'

        Using SQLAlchemy makes it possible to use any DB supported by that
        library. If a DBAPI2 object, only sqlite3 is supported. The user is responsible
        for engine disposal and connection closure for the SQLAlchemy connectable. See
        `here <https://docs.sqlalchemy.org/en/13/core/connections.html>`_
    index_col : str or list of strings, optional, default: None
        Column(s) to set as index(MultiIndex).
    coerce_float : bool, default True
        Attempts to convert values of non-string, non-numeric objects (like
        decimal.Decimal) to floating point, useful for SQL result sets.
    params : list, tuple or dict, optional, default: None
        List of parameters to pass to execute method.  The syntax used
        to pass parameters is database driver dependent. Check your
        database driver documentation for which of the five syntax styles,
        described in PEP 249's paramstyle, is supported.
        Eg. for psycopg2, uses %(name)s so use params={'name' : 'value'}.
    parse_dates : list or dict, default: None
        - List of column names to parse as dates.
        - Dict of ``{column_name: format string}`` where format string is
          strftime compatible in case of parsing string times, or is one of
          (D, s, ns, ms, us) in case of parsing integer timestamps.
        - Dict of ``{column_name: arg dict}``, where the arg dict corresponds
          to the keyword arguments of :func:`pandas.to_datetime`
          Especially useful with databases without native Datetime support,
          such as SQLite.
    columns : list, default: None
        List of column names to select from SQL table (only used when reading
        a table).
    chunksize : int, default None
        If specified, return an iterator where `chunksize` is the number of rows
        to include in each chunk. Note that this argument is only kept for
        compatibility. If a non-none value passed, an error will be reported.
    incremental_index: bool, default False
        Sort RangeIndex if csv doesn't contain index columns.
    test_rows: int, default 5
        The number of rows to fetch for inferring dtypes.
    chunk_size: : int or tuple of ints, optional
        Specifies chunk size for each dimension.
    engine_kwargs: dict, default None
        Extra kwargs to pass to sqlalchemy.create_engine
    incremental_index: bool, default False
        Create a new RangeIndex if csv doesn't contain index columns.
    partition_col : str, default None
        Specify name of the column to split the result of the query. If
        specified, the range ``[low_limit, high_limit]`` will be divided
        into ``n_partitions`` chunks with equal lengths. We do not
        guarantee the sizes of chunks be equal. When the value is None,
        ``OFFSET`` and ``LIMIT`` clauses will be used to cut the result
        of the query.
    num_partitions : int, default None
        The number of chunks to divide the result of the query into,
        when ``partition_col`` is specified.
    low_limit : default None
        The lower bound of the range of column ``partition_col``. If not
        specified, a query will be executed to query the minimum of
        the column.
    high_limit : default None
        The higher bound of the range of column ``partition_col``. If not
        specified, a query will be executed to query the maximum of
        the column.

    Returns
    -------
    DataFrame

    See Also
    --------
    read_sql_table : Read SQL database table into a DataFrame.
    read_sql_query : Read SQL query into a DataFrame.
    """
    return _read_sql(table_or_sql=sql, con=con, index_col=index_col, coerce_float=coerce_float,
                     params=params, parse_dates=parse_dates, columns=columns,
                     engine_kwargs=engine_kwargs, incremental_index=incremental_index,
                     chunksize=chunksize, test_rows=test_rows, chunk_size=chunk_size,
                     partition_col=partition_col, num_partitions=num_partitions,
                     low_limit=low_limit, high_limit=high_limit)


def read_sql_table(table_name, con, schema=None, index_col=None, coerce_float=True,
                   parse_dates=None, columns=None, chunksize=None, test_rows=5,
                   chunk_size=None, engine_kwargs=None, incremental_index=False,
                   use_arrow_dtype=None, partition_col=None, num_partitions=None,
                   low_limit=None, high_limit=None):
    """
    Read SQL database table into a DataFrame.

    Given a table name and a SQLAlchemy connectable, returns a DataFrame.
    This function does not support DBAPI connections.

    Parameters
    ----------
    table_name : str
        Name of SQL table in database.
    con : SQLAlchemy connectable or str
        A database URI could be provided as as str.
        SQLite DBAPI connection mode not supported.
    schema : str, default None
        Name of SQL schema in database to query (if database flavor
        supports this). Uses default schema if None (default).
    index_col : str or list of str, optional, default: None
        Column(s) to set as index(MultiIndex).
    coerce_float : bool, default True
        Attempts to convert values of non-string, non-numeric objects (like
        decimal.Decimal) to floating point. Can result in loss of Precision.
    parse_dates : list or dict, default None
        - List of column names to parse as dates.
        - Dict of ``{column_name: format string}`` where format string is
          strftime compatible in case of parsing string times or is one of
          (D, s, ns, ms, us) in case of parsing integer timestamps.
        - Dict of ``{column_name: arg dict}``, where the arg dict corresponds
          to the keyword arguments of :func:`pandas.to_datetime`
          Especially useful with databases without native Datetime support,
          such as SQLite.
    columns : list, default None
        List of column names to select from SQL table.
    chunksize : int, default None
        If specified, returns an iterator where `chunksize` is the number of
        rows to include in each chunk. Note that this argument is only kept
        for compatibility. If a non-none value passed, an error will be
        reported.
    test_rows: int, default 5
        The number of rows to fetch for inferring dtypes.
    chunk_size: : int or tuple of ints, optional
        Specifies chunk size for each dimension.
    engine_kwargs: dict, default None
        Extra kwargs to pass to sqlalchemy.create_engine
    incremental_index: bool, default False
        Create a new RangeIndex if csv doesn't contain index columns.
    use_arrow_dtype: bool, default None
        If True, use arrow dtype to store columns.
    partition_col : str, default None
        Specify name of the column to split the result of the query. If
        specified, the range ``[low_limit, high_limit]`` will be divided
        into ``n_partitions`` chunks with equal lengths. We do not
        guarantee the sizes of chunks be equal. When the value is None,
        ``OFFSET`` and ``LIMIT`` clauses will be used to cut the result
        of the query.
    num_partitions : int, default None
        The number of chunks to divide the result of the query into,
        when ``partition_col`` is specified.
    low_limit : default None
        The lower bound of the range of column ``partition_col``. If not
        specified, a query will be executed to query the minimum of
        the column.
    high_limit : default None
        The higher bound of the range of column ``partition_col``. If not
        specified, a query will be executed to query the maximum of
        the column.

    Returns
    -------
    DataFrame
        A SQL table is returned as two-dimensional data structure with labeled
        axes.

    See Also
    --------
    read_sql_query : Read SQL query into a DataFrame.
    read_sql : Read SQL query or database table into a DataFrame.

    Notes
    -----
    Any datetime values with time zone information will be converted to UTC.

    Examples
    --------
    >>> import mars.dataframe as md
    >>> md.read_sql_table('table_name', 'postgres:///db_name')  # doctest:+SKIP
    """
    return _read_sql(table_or_sql=table_name, con=con, schema=schema, index_col=index_col,
                     coerce_float=coerce_float, parse_dates=parse_dates, columns=columns,
                     engine_kwargs=engine_kwargs, incremental_index=incremental_index,
                     use_arrow_dtype=use_arrow_dtype, chunksize=chunksize,
                     test_rows=test_rows, chunk_size=chunk_size,
                     partition_col=partition_col, num_partitions=num_partitions,
                     low_limit=low_limit, high_limit=high_limit)


def read_sql_query(sql, con, index_col=None, coerce_float=True, params=None, parse_dates=None,
                   chunksize=None, test_rows=5, chunk_size=None, engine_kwargs=None,
                   incremental_index=None, use_arrow_dtype=None,
                   partition_col=None, num_partitions=None,
                   low_limit=None, high_limit=None):
    """
    Read SQL query into a DataFrame.

    Returns a DataFrame corresponding to the result set of the query
    string. Optionally provide an `index_col` parameter to use one of the
    columns as the index, otherwise default integer index will be used.

    Parameters
    ----------
    sql : str SQL query or SQLAlchemy Selectable (select or text object)
        SQL query to be executed.
    con : SQLAlchemy connectable(engine/connection), database str URI,
        or sqlite3 DBAPI2 connection
        Using SQLAlchemy makes it possible to use any DB supported by that
        library.
        If a DBAPI2 object, only sqlite3 is supported.
    index_col : str or list of strings, optional, default: None
        Column(s) to set as index(MultiIndex).
    coerce_float : bool, default True
        Attempts to convert values of non-string, non-numeric objects (like
        decimal.Decimal) to floating point. Useful for SQL result sets.
    params : list, tuple or dict, optional, default: None
        List of parameters to pass to execute method.  The syntax used
        to pass parameters is database driver dependent. Check your
        database driver documentation for which of the five syntax styles,
        described in PEP 249's paramstyle, is supported.
        Eg. for psycopg2, uses %(name)s so use params={'name' : 'value'}.
    parse_dates : list or dict, default: None
        - List of column names to parse as dates.
        - Dict of ``{column_name: format string}`` where format string is
          strftime compatible in case of parsing string times, or is one of
          (D, s, ns, ms, us) in case of parsing integer timestamps.
        - Dict of ``{column_name: arg dict}``, where the arg dict corresponds
          to the keyword arguments of :func:`pandas.to_datetime`
          Especially useful with databases without native Datetime support,
          such as SQLite.
    chunksize : int, default None
        If specified, return an iterator where `chunksize` is the number of
        rows to include in each chunk. Note that this argument is only kept
        for compatibility. If a non-none value passed, an error will be
        reported.
    incremental_index: bool, default False
        Sort RangeIndex if csv doesn't contain index columns.
    use_arrow_dtype: bool, default None
        If True, use arrow dtype to store columns.
    test_rows: int, default 5
        The number of rows to fetch for inferring dtypes.
    chunk_size: : int or tuple of ints, optional
        Specifies chunk size for each dimension.
    engine_kwargs: dict, default None
        Extra kwargs to pass to sqlalchemy.create_engine
    incremental_index: bool, default False
        Create a new RangeIndex if csv doesn't contain index columns.
    partition_col : str, default None
        Specify name of the column to split the result of the query. If
        specified, the range ``[low_limit, high_limit]`` will be divided
        into ``n_partitions`` chunks with equal lengths. We do not
        guarantee the sizes of chunks be equal. When the value is None,
        ``OFFSET`` and ``LIMIT`` clauses will be used to cut the result
        of the query.
    num_partitions : int, default None
        The number of chunks to divide the result of the query into,
        when ``partition_col`` is specified.
    low_limit : default None
        The lower bound of the range of column ``partition_col``. If not
        specified, a query will be executed to query the minimum of
        the column.
    high_limit : default None
        The higher bound of the range of column ``partition_col``. If not
        specified, a query will be executed to query the maximum of
        the column.

    Returns
    -------
    DataFrame

    See Also
    --------
    read_sql_table : Read SQL database table into a DataFrame.
    read_sql

    Notes
    -----
    Any datetime values with time zone information parsed via the `parse_dates`
    parameter will be converted to UTC.
    """
    return _read_sql(table_or_sql=sql, con=con, index_col=index_col, coerce_float=coerce_float,
                     params=params, parse_dates=parse_dates, engine_kwargs=engine_kwargs,
                     incremental_index=incremental_index, use_arrow_dtype=use_arrow_dtype,
                     chunksize=chunksize, test_rows=test_rows,
                     chunk_size=chunk_size, partition_col=partition_col, num_partitions=num_partitions,
                     low_limit=low_limit, high_limit=high_limit)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from ...actors import FunctionActor
from ...config import options
from ...errors import StorageDataExists
from ...serialize import dataserializer
from ..dataio import ArrowComponentsIO
from .core import StorageHandler, ObjectStorageMixin, BytesStorageIO, \
    DataStorageDevice, wrap_promised, register_storage_handler_cls

try:
    import vineyard
except ImportError:
    vineyard = None
try:
    import pyarrow
except ImportError:
    pyarrow = None


logger = logging.getLogger(__name__)


class VineyardKeyMapActor(FunctionActor):
    @classmethod
    def default_uid(cls):
        return 'w:0:' + cls.__name__

    def __init__(self):
        super().__init__()
        self._mapping = dict()

    def put(self, session_id, chunk_key, obj_id):
        session_chunk_key = (session_id, chunk_key)
        if session_chunk_key in self._mapping:
            raise StorageDataExists(session_chunk_key)
        self._mapping[session_chunk_key] = obj_id

    def get(self, session_id, chunk_key):
        return self._mapping.get((session_id, chunk_key))

    def delete(self, session_id, chunk_key):
        try:
            del self._mapping[(session_id, chunk_key)]
        except KeyError:
            pass

    def batch_delete(self, session_id, chunk_keys):
        for k in chunk_keys:
            self.delete(session_id, k)


class VineyardBytesIO(BytesStorageIO):
    storage_type = DataStorageDevice.VINEYARD

    def __init__(self, vineyard_client, session_id, data_key, data_id, mode='w',
                 nbytes=None, packed=False, compress=None, auto_register=True,
                 pin_token=None, handler=None):
        from .objectholder import SharedHolderActor

        logger.debug('create vineyard bytes IO: mode = %s, packed = %s', mode, packed)

        super().__init__(session_id, data_key, mode=mode, handler=handler)
        self._client = vineyard_client
        self._data_id = data_id
        self._components = None
        self._offset = 0
        self._nbytes = nbytes
        self._holder_ref = self._storage_ctx.actor_ctx.actor_ref(SharedHolderActor.default_uid())
        self._compress = compress or dataserializer.CompressType.NONE
        self._packed = packed
        self._auto_register = auto_register
        self._pin_token = pin_token

        block_size = options.worker.copy_block_size

        if self.is_readable:
            logger.debug('bytes io get: session_id = %s, data_key = %s, data_id = %r, type(data_id) = %r',
                         session_id, data_key, data_id, type(data_id))
            data = self._client.get(data_id)

            self._components = pyarrow.serialize(data, dataserializer.mars_serialize_context()).to_components()
            if packed:
                self._buf = ArrowComponentsIO(
                    self._components, 'r', compress_out=compress, block_size=block_size)
            else:
                raise NotImplementedError('Unknown how to read vineyard values in a unpacked way')
        else:
            raise NotImplementedError

    def __del__(self):
        self._buf = self._components = None

    @property
    def nbytes(self):
        return self._nbytes

    def read(self, size=-1):
        if self._packed:
            return self._buf.read(size)
        else:
            raise NotImplementedError('Unknown how to read vineyard values in a unpacked way')

    def close(self, finished=True):
        if self._closed:
            return
        self._buf = self._components = None
        super().close(finished=finished)


class VineyardHandler(StorageHandler, ObjectStorageMixin):
    storage_type = DataStorageDevice.VINEYARD

    def __init__(self, storage_ctx, proc_id=None):
        StorageHandler.__init__(self, storage_ctx, proc_id=proc_id)
        self._client = vineyard.connect(options.vineyard.socket)
        logger.debug('find mapper ref: %s', VineyardKeyMapActor.default_uid())
        self._mapper_ref = self._storage_ctx.actor_ctx.actor_ref(VineyardKeyMapActor.default_uid())
        logger.debug('find mapper ref done: %s', VineyardKeyMapActor.default_uid())

    def _new_object_id(self, session_id, data_key, data_id):
        logger.debug('mapper put: session_id = %s, data_key = %s, data_id = %r', session_id, data_key, data_id)
        self._mapper_ref.put(session_id, data_key, data_id)

    def _get_object_id(self, session_id, data_key):
        obj_id = self._mapper_ref.get(session_id, data_key)
        logger.debug('mapper get: session_id = %s, data_key = %s, obj_id = %r', session_id, data_key, obj_id)
        if obj_id is None:
            raise KeyError((session_id, data_key))
        return obj_id

    @wrap_promised
    def create_bytes_reader(self, session_id, data_key, packed=False, packed_compression=None,
                            _promise=False):
        data_id = self._get_object_id(session_id, data_key)
        logger.debug('create vineyard bytes reader: data_id = %s', data_id)
        return VineyardBytesIO(self._client, session_id, data_key, data_id, 'r', packed=packed,
                               compress=packed_compression, handler=self)

    @wrap_promised
    def create_bytes_writer(self, session_id, data_key, total_bytes, packed=False,
                            packed_compression=None, auto_register=True, pin_token=None,
                            _promise=False):
        raise NotImplementedError('vineyard bytes writer hasn\'t been implemented')

    @wrap_promised
    def get_objects(self, session_id, data_keys, serialize=False, _promise=False):
        data_ids = [self._get_object_id(session_id, data_key) for data_key in data_keys]
        return self._client.get_object(data_ids)

    @wrap_promised
    def put_objects(self, session_id, data_keys, objs, sizes=None, shapes=None,
                    serialize=False, pin_token=None, _promise=False):
        sizes, shapes = [], []
        for data_key, obj in zip(data_keys, objs):
            if isinstance(obj, pyarrow.SerializedPyObject):
                obj = obj.deserialize(dataserializer.mars_serialize_context())
            data_id, size, shape = self._client.put_object(obj)
            sizes.append(size)
            shapes.append(shape)
            self._new_object_id(session_id, data_key, data_id)
        self.register_data(session_id, data_keys, sizes, shapes)

    def load_from_bytes_io(self, session_id, data_keys, src_handler, pin_token=None):
        def _read_serialized(reader):
            with reader:
                return reader.get_io_pool().submit(reader.read).result()

        def _fallback(*_):
            return self._batch_load_objects(
                session_id, data_keys,
                lambda k: src_handler.create_bytes_reader(session_id, k, _promise=True).then(_read_serialized),
                serialize=True
            )

        return self.transfer_in_runner(session_id, data_keys, src_handler, _fallback)

    def load_from_object_io(self, session_id, data_keys, src_handler, pin_token=None):
        def _fallback(*_):
            return self._batch_load_objects(
                session_id, data_keys,
                lambda k: src_handler.get_objects(session_id, k, _promise=True), batch_get=True)

        return self.transfer_in_runner(session_id, data_keys, src_handler, _fallback)

    def delete(self, session_id, data_keys, _tell=False):
        data_ids = [self._get_object_id(session_id, data_key)
                    for data_key in data_keys]
        self._client.delete(data_ids, deep=True)
        self.unregister_data(session_id, data_keys, _tell=_tell)


register_storage_handler_cls(DataStorageDevice.VINEYARD, VineyardHandler)
